// MD_BUFFOBJ.JS
//
// Author:	Poundboy
// Email:	poundboy@megapathdsl.net
// Version:	1.02
//
// Description:
//
// The following routines implement a buff object which handles buffing of all players and items.
// This object has the following features:
//
// Buffs can be set for self, items, and others
//
// Algorithm uses estimated expiration times to cast next buff instead of detecting spell expirations.
// While this may seen inefficient, we do this for a couple reasons:
// 1) I can only get the exact time left on a spell for self life/creature buffs.  Spells on items
//    and other players I can't get the time left.
// 2) Since the spell economy was removed all spells last a fixed amount of time
// 3) For buffs that we want to remain active at all times (have a safety margin), I can't wait for the
//    spell expire message since the buff would be briefly inactive.
//
//
// Requirements:
//
// Requires the following libraries: MD_UTIL.JS, MD_LOGOBJ.JS, MD_CAST.JS
//
// Usage:
//
// A default object of the name Buff is created automatically when this file is included in a project.
// This object should be used for all self and item buffs.
//
// Object Constructor:
//
// myObj = new BuffObj(Debug_Flag, opm)
//
// where Debug_Flag is true or false, and opm indicates the destination of the debug output.
//
// Methods:
//
//
// BuffInit()				Goes through active spells on character and updates
//					estimated expiration time of any matching spells.  Should
//					only be called once
//
// Check()				Returns count of buffs expired.
//
// Cast(castmax)			Casts up to 'castmax' spells.  Returns true if no spells
//					cast, false if spell cast done.   For time critical macros
//					you should only cast one buff maximum at a time and check
//					for critical events between calls to Buff.Cast(). Note that
//					a failed cast attempt is counted as a cast.
//
// ExpireUncastItem()			This will set all uncast item spells to unexpired.
//					Used in Warbot macro to enable user to force casting
//					of item spells immediately.
//
// ExpireAll()				This will set all spells to expired.  Used in Warbot macro
//					to force casting of all spells immediately
//
// Extend(time_ms)			Since we use absolute times for estimating buff expirations,
//					anytime we logoff/on under program control (for example to hide
//					from other players) we must adjust the estimated expiration
//					times by the length of the logout period.
//
// ShowBuffs()				List out all buffs enabled.  For debugging
//
// SetDebug(boolean,opm)		Enable/Disable debug mode.
//
//
// Modification History
//
// Who	Date	  Version	Description
//
// PB	06/23/02  1.01		Initial version
//
// PB	09/20/02  1.02		1) For self-buffs, no longer have to estimate spell expiration; can
//				   get info from skapi.cospell.  Also useful when creature/life buffs
//				   already cast when macro starts
//
//				2) For item buffs, still can't figure out duration, but now we check
//				   item for active buff before casting first time.   After 1st cast
//				   we estimate spell expiration as before so spell can be recast
//				   before expiration (if safety factor specified)
//
//				3) Get rid of detection of spells cast on us from other players.
//				   Simplifies spell casting and buff objects
//
// **************************************************************************************************
// Global constants
//

buffMaxItemSpells = 30;	// Max # of item spells on buff list

// Define constants that define buff types
typeSelf	= 1;	// Creature/Life self buff
typeItem	= 2;	// Item buff on an object
typeOther	= 3;	// Creature/Life buff on another player

buffType	= new Array();
buffTypeCre	= 1;
buffTypeLife	= 2;
buffTypeItem	= 3;
buffTypeFirst	= 1;
buffTypeLast	= 3;
buffTypeText	= new Array();
buffTypeText[buffTypeCre]	= "Creature";
buffTypeText[buffTypeLife]	= "Life";
buffTypeText[buffTypeItem]	= "Item";

equipClassArmor	  = 1;
equipClassWeapon  = 2;
equipClassUnknown = 3;
equipClassText	= new Array();
equipClassText[equipClassArmor]	  = "Armor";
equipClassText[equipClassWeapon]  = "Weapon";
equipClassText[equipClassUnknown] = "Unknown";

// Now define default Eqm by equipment class
equipDefEqm	= new Array();
equipDefEqm[equipClassArmor]	= eqmShield | eqmHead | eqmHands | eqmFeet | eqmChestOuter | eqmAbdomenOuter |
				  eqmUpperArmsOuter | eqmLowerArmsOuter | eqmUpperLegsOuter | eqmLowerLegsOuter;	     
equipDefEqm[equipClassWeapon]	= eqmFocusWeapon | eqmWeapon | eqmRangedWeapon;

// Define filters for equipment types
equipEqmFilter	= new Array();
equipEqmFilter[equipClassArmor]	= eqmShield | eqmHead | eqmHands | eqmFeet | eqmChestOuter | eqmAbdomenOuter |
				  eqmUpperArmsOuter | eqmLowerArmsOuter | eqmUpperLegsOuter | eqmLowerLegsOuter |
				  eqmChestUnder | eqmUpperLegsUnder;
equipEqmFilter[equipClassWeapon]= eqmFocusWeapon | eqmWeapon | eqmRangedWeapon;
						
equipActiveTrue	= true;
equipActiveFalse= false;

// Defaults for buffs when parameters not specified
buffDefLevel	= 7;
buffDefPrio	= 100;
buffDefSafety	= 180;		// 180 seconds
buffDefRandom	= true;

// Miscellaneous constants
buffHoldDown	= 4000;		// Don't recast a buff within 4000ms (4 seconds) of previous cast attempt.
				// This is needed since sometimes the program doesn't detect a successful
				// cast spell for a few seconds so this stops the buff from being recast
				// needlessly.

// Global variables.  Can't put most of these in objects since events fire outside of object context
BuffDebug	= false;	// Set to true when Buff object logging desired
BuffDebug2	= false;	// Extended debug info, set manually by editing this file
BuffLog		= new LogObj();	// Create log object for comp routines

BuffPrevAco	= null;		// Used to determine if item spell is being added to same
    				// target as previous one.  If so then no need to assess the item.


function BuffInfoRec(Line) {
    if (BuffDebug2) BuffLog.OutD("BuffInfoRec: Begin");

    // Name, Type, Class ,L1, L2, L3, L4, L5, L6, L7, L1, L2, L3, L4, L5, L6, L7
    // 0     1     2      3   4   5   6   7   8   9   10  11  12  13  14  15  16
    Split = Line.split(",");
    this.Name		= Util.Trim(Split[0]);
    this.Type		= parseInt(Split[1]);
    this.Class		= parseInt(Split[2]);
    this.Self 		= new Array();
    this.Self[0]	= 0;
    for (var x=1; x<8; x++) this.Self[x] = parseInt(Split[2+x]);
    this.Other 	= new Array();
    this.Other[0]	= 0;
    for (var x=1; x<8; x++) this.Other[x] = parseInt(Split[9+x]);

    // Now initialize default values for user settable buff information.
    // These will be over written by users buff config file (if defined).
    this.Enabled	= false;	 	// This buff enabled?
    this.Level		= buffDefLevel;	 	// Minimum level buff that must be active (could be self or other)
    this.Priority	= buffDefPrio;	 	// Priority of buff.  Higher ones cast first.
    this.Safety		= buffDefSafety*1000; 	// Time before spell buff expires that buff should be recast.
    this.Randomize	= buffDefRandom; 	// If true, randomize second cast of buff to spread them out.
    this.RandomTime	= 0;			// Additional safety time to add to spell for 2nd cast if random
    						// flag true. Will be filled in after 1 cast.

    // Set default equipment mask for this spell. Only valid for armor item spells.
    this.Eqm		= equipDefEqm[this.Class];

    // Define some internal variables for this buff spell. Only used for creature and life spells.
    // Item spell buff statuses stored in equipment list
    this.Count		= 0;			// # of time this buff has been cast
    this.LastCast	= 0;			// Last time this spell was cast
    this.EstExp 	= 0;			// Assume spell is expired at begining
    this.Force		= false;		// Set to true when we want to force casting of a buff
    this.SpellObj	= skapi.SpellInfoFromSpellid(this.Self[this.Level]);

    // Now determine the family for this spell group
    this.Family		= this.SpellObj.family;

   if (BuffDebug2) BuffLog.OutD("BuffInfoRec: End");
} // end func

function BuffPermStatsRec() {
    if (BuffDebug2) BuffLog.OutD("BuffPermStatsRec: Begin");
    this.Enabled	= false;	 	// This buff enabled?
    this.Level		= buffDefLevel;	 	// Minimum level buff that must be active (could be self or other)
    this.Priority	= buffDefPrio;	 	// Priority of buff.  Higher ones cast first.
    this.Safety		= buffDefSafety*1000; 	// Time before spell buff expires that buff should be recast.
    this.Randomize	= buffDefRandom; 	// If true, randomize second cast of buff to spread them out.
    this.SpellObj	= null;

    this.Count		= 0;
    this.LastCast	= 0;
    this.EstExp		= 0;
    this.Force		= false;		// Set to true when we want to force casting of a buff

    if (BuffDebug2) BuffLog.OutD("BuffPermStatsRec: End");
} // end func

function BuffTempStatsRec() {
    if (BuffDebug2) BuffLog.OutD("BuffTempStatsRec: Begin");
    this.Count		= 0;
    this.LastCast	= 0;
    this.EstExp		= 0;
    this.Force		= false;		// Set to true when we want to force casting of a buff
    if (BuffDebug2) BuffLog.OutD("BuffTempStatsRec: End");
} // end func


function BuffPermEquipRec(Aco, Class) {
    if (BuffDebug2) BuffLog.OutD("BuffPermEquipRec: Begin");
    this.Aco	  = Aco;	// aco of equipment
    this.Oid	  = Aco.oid;	// oid of equipment
    this.Name	  = Aco.szName;	// Name of equipment
    this.Class	  = Class;	// Type of equipment (armor or weapon).
    this.RendType = 0;		// If equipment is a wand, then we store it's render type here (if it has one)
    this.Exists	  = true;	// Set to false if no longer in inventory
    this.Active	  = false;	// Active flag.  True if item being worn, false otherwise

    // Now setup buff stat record for each possible item buff
    this.Stats	= new Array();
    for (var x=0; x<buffMaxItemSpells; x++) this.Stats[x] = new BuffPermStatsRec();

    if (BuffDebug2) BuffLog.OutD("BuffPermEquipRec: End");
} // endif

function BuffTempEquipRec(Aco, Class) {
    if (BuffDebug2) BuffLog.OutD("BuffTempEquipRec: Begin");
    this.Aco	  = Aco;	// aco of equipment
    this.Oid	  = Aco.oid;	// oid of equipment
    this.Name	  = Aco.szName;	// Name of equipment
    this.Class	  = Class;	// Type of equipment (armor or weapon).
    this.RendType = 0;		// If equipment is a wand, then we store it's render type here (if it has one)
    this.Active	  = true;	// Active flag.  True if item being worn, false otherwise

    // Now setup buff stat record for each possible item buff
    this.Stats	= new Array();
    for (var x=0; x<buffMaxItemSpells; x++) this.Stats[x] = new BuffTempStatsRec();

    if (BuffDebug2) BuffLog.OutD("BuffTempEquipRec: End");
} // endif

function BuffRangeRec(Begin,End) {
    if (BuffDebug2) BuffLog.OutD("BuffRangeRec: Begin");
    this.Begin	= Begin;
    this.End	= End;
    if (BuffDebug2) BuffLog.OutD("BuffRangeRec: End");
} // endfunc


function IdxObj(InfoIdx, Level) {
    this.InfoIdx = InfoIdx;
    this.Level   = Level;
} // end func


// Contains a list of buffs
function BuffObj(debugflg) {
    if (BuffDebug) BuffLog.OutD("BuffObj: Begin");
    if ((debugflg==undefined) || (debugflg==null)) debugflg = false;
    BuffDebug = debugflg;   // Outputs debugging info if set

    // Initialize buff list.  This list contains all possible buffs
    this.Info		= new Array();			// Buff list
    this.InfoCnt	= 0;    			// # of buffs in list
    this.IdxFrmSpell	= new Array();			// Converts Skunkworks spell id to BuffList index

    // Earliest time a buff will expire.  0 means unknown.
    this.NextExp	= 0;

    // Holds equipment defined in the equipment data file.
    this.PermEquip	= new Array();
    this.PermEquipCnt	= 0;

    // Initialize temporary equipment list.  Used to keep track of equipment we are wearing that is not
    // defined in the permanent equipment list.
    this.TempEquip	= new Array();
    this.TempEquipCnt	= 0;

    // Init array that maps beginning and ending Info indexes for each buff and class type
    this.Range		= new Array();	// Range for each buff type
    for (var x=0; x<4; x++) this.Range[x] = new BuffRangeRec();
    this.ClassRange	= new Array();	// Range for each class type
    for (var x=0; x<4; x++) this.ClassRange[x] = new BuffRangeRec();
    this.ItemCnt	= 0;		// # of item buffs

    // File context
    this.FileSysObj	= new ActiveXObject("Scripting.FileSystemObject");
    this.FileName	= 0;		// Last filename index used
    
    // Flag that indicated whether item buff logging is enabled or not
    this.LogBuffs	= false;	// Item buff logging is enabled or not?
    this.LogFile	= "";		// File to log buffs to
    this.LogFileObj	= null;		// Log file object.  Used to write and read records

    // Flags set when special markers detected in buff config file
    this.BOF		= false;
    this.EOF		= false;
    
    // If FORCE flag set on a spell, then buff is recast unless it was
    // cast within this many milliseconds
    this.ForceIgnoreInt	= 45*1000;	// 45*1000 = 45 seconds

    // Public Methods
    this.SetDebug	  = Buff_SetDebug;		// Set debug mode
    this.Init		  = Buff_Init;			// Load static buff list from file
    this.Save		  = Buff_Save;			// Save user defined buff settings to a file
    this.Load		  = Buff_Load;			// Load user defined buff settings from a file
    this.Show		  = Buff_Show;			// Show buffs
    this.Enable		  = Buff_Enable;		// Enable a buff
    this.Disable	  = Buff_Disable;		// Disable a buff
    this.Reset		  = Buff_Reset;			// Reset all buffs to default settings
    this.Lookup		  = Buff_Lookup;		// Return a buffs index given its name
    this.SetLevel	  = Buff_SetLevel;		// Set minimum spell level for a buff
    this.SetPriority	  = Buff_SetPriority;		// Set Priority property in a BuffInfoRec
    this.SetSafety	  = Buff_SetSafety;		// Set Safety property in a BuffInfoRec
    this.SetRandomize	  = Buff_SetRandomize;		// Set Randomize property in a BuffInfoRec
    this.SetEqm		  = Buff_SetEqm;		// Set Eqm property in a BuffInfoRec
    this.UpdateExpire	  = Buff_UpdateExpire;		// Update spell expirations from currently active spell list
    this.CheckEquip	  = Buff_CheckEquip;		// Updates equipment list with current items we are wearing
    this.ResetCreLife	  = Buff_ResetCreLife;		// Reset cast count on creature and life spells
    this.ForceUncastItem  = Buff_ForceUncastItem;	// Force all uncast item buffs
    this.ForceAll	  = Buff_ForceAll;		// Force all spells to be cast
    this.Cast		  = Buff_Cast;			// Cast one buff
    this.Extend		  = Buff_Extend;		// Extend all buffs by given time
    this.ShowBuffs	  = Buff_ShowBuffs;		// for debugging mainly
    this.LoadEquip	  = Buff_LoadEquip;		// Load user defined equipment and spells from a file
    this.ResetPermEquip	  = Buff_ResetPermEquip;	// Zero out permanent equipment list
    this.ProcItemLog	  = Buff_ProcItemLog;		// Process list of item buffs we wrote to file.
    this.SetBuffLog	  = Buff_SetBuffLog;		// Turn item buff logging on or off
    this.LogItemBuff	  = Buff_LogItemBuff;		// Write item buff to file so we can compute expiration
    							// time of item buffs after macro is restarted.
    if (BuffDebug) BuffLog.OutD("BuffObj: End");
} // end func BuffObj


function Buff_SetDebug(debugflag,opm) {
    if (BuffDebug) BuffLog.OutD("Buff_SetDebug: Begin");
    BuffDebug = debugflag;
    if (opm!=undefined) BuffLog.SetOpm(opm);
    if (BuffDebug) BuffLog.OutD("Buff_SetDebug: End");
    return true;
} // end func


function Buff_Init(inputFile) {
    if (BuffDebug) BuffLog.OutD("Buff_Init: Begin");
    var TextFileObj = this.FileSysObj.OpenTextFile(inputFile, 1, false);

    // Read the first line which is the header record and discard
    this.Line1 = TextFileObj.ReadLine();
    var Type = 0;
    var Class = 0;

    // Read rest of records in file until EOF
    while (!(TextFileObj.AtEndOfStream)) {
	// now read lines until EOF
	var Line = TextFileObj.ReadLine();
	this.Info[this.InfoCnt] = new BuffInfoRec(Line);
	for (var x=1; x<8; x++) {
	    this.IdxFrmSpell[this.Info[this.InfoCnt].Self[x]]  = new IdxObj(this.InfoCnt, x);
	    this.IdxFrmSpell[this.Info[this.InfoCnt].Other[x]] = new IdxObj(this.InfoCnt, x);
	} // endfor

	if (Type != this.Info[this.InfoCnt].Type) {
	    // Update begin & end ranges for buff type
	    if (Type>0) this.Range[Type].End = this.InfoCnt-1;
	    Type = this.Info[this.InfoCnt].Type;
	    this.Range[Type].Begin = this.InfoCnt;
	} // endif
	if (Class != this.Info[this.InfoCnt].Class) {
	    // Update begin & end ranges for buff class
	    if (Class>0) this.ClassRange[Class].End = this.InfoCnt-1;
	    Class = this.Info[this.InfoCnt].Class;
	    this.ClassRange[Class].Begin = this.InfoCnt;
	} // endif
	this.InfoCnt++;		// Increment count of buffs
    } // end while
    this.Range[Type].End = this.InfoCnt-1;
    this.ClassRange[Class].End = this.InfoCnt-1;

    this.ItemCnt = (this.Range[buffTypeItem].End - this.Range[buffTypeItem].Begin) + 1;

    // Close file
    TextFileObj.Close();
   if (BuffDebug) BuffLog.OutD("Buff_Init: End");
} // endfunc


function Buff_Save(fileNum) {
    if (BuffDebug) BuffLog.OutD("Buff_Save: Begin");

    // Build the output filename from the filenumber.  Note filenum 0 is a special filename.
    if ((fileNum == undefined) || (fileNum == null)) fileNum = 0;
    if (fileNum == 0) {
    	var Post = "def" + fileCfgExt;
    } else {
    	var Post = fileNum + fileCfgExt;
    	if (fileNum < 10) Post = "0" + Post;
    } // endif
    this.FileName = CharCfgDir + "\\" + fileBuffCfg + Post;

    var SavedCount = 0;

    // Now create the file
    BuffFile = this.FileSysObj.OpenTextFile(this.FileName, ForWriting, true);	// Create config file
    BuffFile.WriteLine("BOF, Buff config file created " + Util.FormatDate(0,false));
    for (var x=0; x<this.InfoCnt; x++) {
	with (this.Info[x]) {
	    if (Enabled) SavedCount++;
	    BuffFile.WriteLine(Name + "," + Enabled + "," + Level + "," + Priority + "," +
	    		       Safety + "," + Randomize + "," + Eqm);
	} // end with
    } // end for

    // Now write special EOF marker
    BuffFile.WriteLine("EOF");

    // All done, close file
    BuffFile.Close();

    if (BuffDebug) BuffLog.OutD("Buff_Save: End");
    return SavedCount;
} // endif

function Buff_Load(fileNum) {
    // Load buff settings from a file
    if (BuffDebug) BuffLog.OutD("Buff_Load: Begin");

    // Build the output filename from the filenumber.  Note filenum 0 is a special filename.
    if ((fileNum == undefined) || (fileNum == null)) fileNum = 0;
    if (fileNum == 0) {
    	var Post = "def" + fileCfgExt;
    } else {
    	var Post = fileNum + fileCfgExt;
    	if (fileNum < 10) Post = "0" + Post;
    } // endif
    this.FileName = CharCfgDir + "\\" + fileBuffCfg + Post;

    // See if file exists
    if(!this.FileSysObj.FileExists(this.FileName)) return -1;

    // Clear out current buffs
    this.Reset();

    // Init misc variables
    var LoadCount = 0;
    this.BOF = false;
    this.EOF = false;

    // Open file
    var TextFileObj = this.FileSysObj.OpenTextFile(this.FileName, 1, false);

    while (!(TextFileObj.AtEndOfStream)) {
	// now read lines until EOF
	var Line  	= TextFileObj.ReadLine();
	var Split 	= Line.split(",");
	var Name	= Util.Trim(Split[0]);

	// See if line is a special marker
	if (Name == "BOF") {
	    this.BOF = true;
	    continue;
	} // endif
	if (Name == "EOF") {
	    this.EOF = true;
	    break;
	} // endif

	if (Util.Trim(Split[1]) == "true") var Enabled = true; else var Enabled = false;
	var Level	= parseInt(Split[2]);
	var Priority	= parseInt(Split[3]);
	var Safety	= parseInt(Split[4]);
	if (Util.Trim(Split[5]) == "true") var Randomize = true; else var Randomize = false;
	var Eqm		= parseInt(Split[6]);

	var BuffIndex = this.Lookup(Name);
	if (BuffIndex < 0) {
	    BuffLog.OutD("Unknown buff name [" + Name + "]", true);
	    continue;
	} // endif

	this.Info[BuffIndex].Enabled	= Enabled;
	this.Info[BuffIndex].Level	= Level;
	this.Info[BuffIndex].Priority	= Priority;
	this.Info[BuffIndex].Safety	= Safety;
	this.Info[BuffIndex].Randomize	= Randomize;
	this.Info[BuffIndex].Eqm	= Eqm;
	this.Info[BuffIndex].SpellObj	= skapi.SpellInfoFromSpellid(this.Info[BuffIndex].Self[Level]);

	if (Enabled && !Cast.Know(this.Info[BuffIndex].Self[Level])) {
	    this.Info[BuffIndex].Enabled = false;
	    Chat.Screen("You do not know " + this.Info[BuffIndex].Name + " Self " +
		    	this.Info[BuffIndex].Level + ", buff disabled.");
	} // endif
	if (Enabled) LoadCount++;	// Increment count of enabled buffs read
    } // end while
    
    TextFileObj.Close();

    if (!this.EOF) {
	Chat.Screen("WARNING!!! End of file marker not detected in buff config file!!!", cmcYellow);
	Chat.Screen("Buff file is either corrupt or did not write correctly last save operation!", cmcYellow);
	Chat.Screen("Make sure you buffs are setup correctly!", cmcYellow);
    } // endif

    if (BuffDebug) BuffLog.OutD("Buff_Load: End, LoadCount [" + LoadCount + "]");
    return LoadCount;
} // endif


function Buff_Show(BuffType, Class) {
    if ((BuffType==undefined) || (BuffType==null)) BuffType = 0;
    if ((Class==undefined)    || (Class==null))    Class    = 0;

    if (BuffType==0) {
	var Begin = buffTypeFirst;
	var End   = buffTypeLast;
    } else {
    	var Begin = BuffType;
    	var End   = BuffType;
    } // endif

    var Header = false;
    for (var x=this.Range[Begin].Begin; x<=this.Range[End].End; x++) {
    	if ((Class>0) && (this.Info[x].Class != Class)) continue;

    	// Show enabled buffs
	with (this.Info[x]) {
	    if (Enabled) {
		if (!Header) {
		    Header = true;
		    Chat.Screen(" ");
		    Chat.Screen(Util.Width("Buff",3.0)    + Util.Width("Level",0.75) +
		    	        Util.Width("Safety",0.75) + Util.Width("Prio",0.75) + "Random");
		} // endif

		// Print info
		Chat.Screen(Util.Width(Name,3.0)     + Util.Width(Level + " ",0.75) +
			    Util.Width((Safety/1000) + " ",0.75) + Util.Width(Priority + " ",0.75) + Randomize);
	    } // endif
	} // end with
    } // endfor
    return Header;	// returns false if no buffs found, true if one or more enabled
} // endfunc


function Buff_Enable(BuffIndex, Level, Priority, Safety, Randomize, Eqm) {
    if (BuffDebug) BuffLog.OutD("Buff_Enable: Begin");
    this.Info[BuffIndex].Enabled = true;
    if ((Level     != undefined) && (Level     != null)) this.Info[BuffIndex].Level	= Level;
    if ((Priority  != undefined) && (Priority  != null)) this.Info[BuffIndex].Priority	= Priority;
    if ((Safety    != undefined) && (Safety    != null)) this.Info[BuffIndex].Safety	= Safety;
    if ((Randomize != undefined) && (Randomize != null)) this.Info[BuffIndex].Randomize	= Randomize;
    if ((Eqm       != undefined) && (Eqm       != null)) this.Info[BuffIndex].Eqm	= Eqm;

    // Now check to see if we know the spell we just enabled.  If not disable it and warn the user
    var SpellId = this.Info[BuffIndex].Self[this.Info[BuffIndex].Level];
    if (!Cast.Know(SpellId)) {
    	this.Info[BuffIndex].Enabled = false;
    	var SpellObj = skapi.SpellInfoFromSpellid(SpellId);
    	BuffLog.OutDM("Buff_Enable", "WARNING!!! Spell " + SpellObj.szName + " not known, buff disabled.", true, cmcYellow);
    } // endif

    // Update spell object for this buff
    this.Info[BuffIndex].SpellObj = skapi.SpellInfoFromSpellid(this.Self[this.Info[BuffIndex].Level]);

    if (BuffDebug) BuffLog.OutD("Buff_Enable: End");
} // end func

function Buff_Disable(BuffIndex) {
    if (BuffDebug) BuffLog.OutD("Buff_Disable: Begin");
    this.Info[BuffIndex].Enabled = false;
    if (BuffDebug) BuffLog.OutD("Buff_Disable: End");
} // end func

function Buff_Reset() {
    if (BuffDebug) BuffLog.OutD("Buff_Reset: Begin");
    this.NextExp = 0;
    for (var x=0; x<this.InfoCnt; x++) {
    	this.Info[x].Enabled	= false;
    	this.Info[x].Level	= buffDefLevel;
    	this.Info[x].Priority	= buffDefPrio;
    	this.Info[x].Safety	= buffDefSafety*1000;
    	this.Info[x].Randomize	= buffDefRandom;
    	this.Info[x].Eqm	= equipDefEqm[this.Info[x].Class];
    } // endfor
    if (BuffDebug) BuffLog.OutD("Buff_Reset: End");
} // end func

function Buff_Lookup(BuffName) {
    if (BuffDebug) BuffLog.OutD("Buff_Lookup: Begin");
    for (var x=0; x<this.InfoCnt; x++) if (this.Info[x].Name == BuffName) break;
    if (x==this.InfoCnt) x = -1;
    if (BuffDebug) BuffLog.OutD("Buff_Lookup: End, x = [" + x + "]");
    return x;
} // end func

function Buff_SetLevel(BuffIndex, Level) {
    if (BuffDebug) BuffLog.OutD("Buff_SetLevel: Begin");
    this.Info[BuffIndex].Level = Level;
    // If buff is enabled, then make sure we know the new level.
    if (this.Info[BuffIndex].Enabled) {
	var SpellId = this.Info[BuffIndex].Self[Level];
	if (!Cast.Know(SpellId)) {
    	    this.Info[BuffIndex].Enabled = false;
    	    var SpellObj = skapi.SpellInfoFromSpellid(SpellId);
    	    BuffLog.OutDM("Buff_Enable", "WARNING!!! Spell " + SpellObj.szName + " not known, buff disabled.",
    	    		  true, cmcYellow);
	} // endif
    } // endif
    // We know the spell, update spell object (used when casting this buff)
    this.Info[BuffIndex].SpellObj = skapi.SpellInfoFromSpellid(this.Self[Level]);
    if (BuffDebug) BuffLog.OutD("Buff_SetLevel: End");
} // end func

function Buff_SetPriority(BuffIndex, Priority) {
    if (BuffDebug) BuffLog.OutD("Buff_SetPriority: Begin");
    this.Info[BuffIndex].Priority = Priority;
    if (BuffDebug) BuffLog.OutD("Buff_SetPriority: End");
} // end func

function Buff_SetSafety(BuffIndex, Safety) {
    if (BuffDebug) BuffLog.OutD("Buff_SetSafety: Begin");
    this.Info[BuffIndex].Safety = Safety*1000;
    if (BuffDebug) BuffLog.OutD("Buff_SetSafety: End");
} // end func

function Buff_SetRandomize(BuffIndex, Randomize) {
    if (BuffDebug) BuffLog.OutD("Buff_SetRandomize: Begin");
    this.Info[BuffIndex].Randomize = Randomize;
    if (BuffDebug) BuffLog.OutD("Buff_SetRandomize: End");
} // end func

function Buff_SetEqm(BuffIndex, Eqm) {
    if (BuffDebug) BuffLog.OutD("Buff_SetEqm: Begin");
    this.Info[BuffIndex].Eqm = Eqm;
    if (BuffDebug) BuffLog.OutD("Buff_SetEqm: End");
} // end func

function Buff_UpdateExpire() {
    // Update creature/life buff expirations based on currently active spells
    if (BuffDebug) BuffLog.OutD("Buff_UpdateExpire: Begin");

    // First blank out the estimated expiration times for all life and creature spells
    for (var x=this.Range[buffTypeCre].Begin; x<=this.Range[buffTypeLife].End; x++) this.Info[x].EstExp = 0;

    // Go through currently active spells on our character.  Update the estimated expiration time
    // for spells that we are tracking.  Note that since there may be multiple spells of the same id
    // on us (different layers), we only update the expiration time if the new time is greater than
    // the current expiration time.
    for (var x=0; x < skapi.cospell.Count; x++) {
    	var SpellObj = skapi.cospell(x);
    	var SpellId = SpellObj.spellid;
	if (BuffDebug2) BuffLog.OutD("Buff_UpdateExpire: Checking active spell [" + SpellId + "] Name [" +
	    			    SpellObj.szName + "] Layer [" + SpellObj.layer + "]");

	if (this.IdxFrmSpell[SpellId] == undefined) {
	    if (BuffDebug2) BuffLog.OutD("Buff_UpdateExpire: We are not tracking this spell");
	    continue;
	} // endif

	var InfoIndex  = this.IdxFrmSpell[SpellId].InfoIdx;
	var SpellLevel = this.IdxFrmSpell[SpellId].Level;
	
	// Next see if this buff is enabled for this InfoIndex
	if (BuffDebug2) BuffLog.OutD("Buff_UpdateExpire: InfoIndex [" + InfoIndex + "]");
	if (!this.Info[InfoIndex].Enabled) {
	    if (BuffDebug2) BuffLog.OutD("Buff_UpdateExpire: Buff is disabled, rejected");
	    continue;
	} // endif

	// If spell level of active spell to low for this buff, then ignore it
	if (SpellLevel < this.Info[InfoIndex].Level) {
	    if (BuffDebug2) BuffLog.OutD("Buff_UpdateExpire: Active spell too low, rejected");
	    continue;
	} // endif
    
	// Set spell exp to be current time + time left on spell - (safety margin or random time)
	var Remain = SpellObj.csecRemain;
	if (BuffDebug2) BuffLog.OutD("Buff_UpdateExpire: Remaining time [" + Remain + "]")
	var CurTime = new Date();
	NewExp = CurTime.getTime() + (Remain*1000);
	if ((this.Info[InfoIndex].Count == 1) && (this.Info[InfoIndex].Randomize)) {
	    NewExp -= this.Info[InfoIndex].RandomTime;
	    if (BuffDebug2) BuffLog.OutD("Buff_UpdateExpire: Subtracting random time.  This will be second cast.");
	} else {
	    NewExp -=  this.Info[InfoIndex].Safety;
	    if (BuffDebug2) BuffLog.OutD("Buff_UpdateExpire: Subtracting safety margin");
	} // endif
	if (BuffDebug2) BuffLog.OutD("Buff_UpdateExpire: Calc exp on this spell [" + new Date(NewExp) +"]");
    
	// If new expiration is greater than the old one, then replace it
	if (NewExp > this.Info[InfoIndex].EstExp) {
	    this.Info[InfoIndex].EstExp = NewExp;
	    if (BuffDebug2) BuffLog.OutD("Buff_UpdateExpire: Setting new exp");
	} else {
	    if (BuffDebug2) BuffLog.OutD("Buff_UpdateExpire: Current exp time [" +
					  new Date(this.Info[InfoIndex].EstExp) + "] further in future.");
	} // endif
    } // endfor (x)

    if (BuffDebug) BuffLog.OutD("Buff_UpdateExpire: End");
} // end func

function Buff_CheckEquip() {
    if (BuffDebug) BuffLog.OutD("Buff_CheckEquip: Begin");
    // First clear the active flag for all current items on equipment list. Also
    // regenerate ACOs as they may be stale due to a logoff/login sequence

    // The value returned from this function is the render type of the wand we holding  
    var ReturnValue = 0;

    // First deal with the temporary list
    var x=0;
    while (x<this.TempEquipCnt) {
	this.TempEquip[x].Active = false;
	this.TempEquip[x].Aco = skapi.AcoFromOid(this.TempEquip[x].Oid);
	// If aco does not exist anymore, then remove it from list
	if ((this.TempEquip[x].Aco == null) || !this.TempEquip[x].Aco.fExists) {
	    // Remove from list by copying last item to current position
	    if (BuffDebug) BuffLog.OutD("Buff_ChkEquip: Equipment [" + this.TempEquip[x].Name + "/" +
	    				 Util.Hex(this.TempEquip[x].Oid) + "] no longer exists");
	    this.TempEquipCnt--;
	    if (x<this.TempEquipCnt) this.TempEquip[x] =  this.TempEquip[this.TempEquipCnt];
	} else {
	    x++;
	} // endif
    } // end while

    // Now process the permanent list
    for (var x=0; x<this.PermEquipCnt; x++) {
	this.PermEquip[x].Active = false;
	this.PermEquip[x].Exists = true;
	this.PermEquip[x].Aco = skapi.AcoFromOid(this.PermEquip[x].Oid);
	// If aco does not exist anymore, or its not in inventory or being worn, then disable buffing of it.
	if ((this.PermEquip[x].Aco == null) || !this.PermEquip[x].Aco.fExists ||
	    (!this.PermEquip[x].Aco.fInInventory && (this.PermEquip[x].Aco.eqmWearer == eqmNil))) {
	    // Mark it as disabled
	    this.PermEquip[x].Exists = false;
	    if (BuffDebug) BuffLog.OutD("Buff_ChkEquip: Perm Equipment [" + this.PermEquip[x].Name + "/" +
	    				 Util.Hex(this.PermEquip[x].Oid) + "] no longer in inventory");
	} // endif
    } // end for

    // Build filter to return items we are wearing
    var ItemAcf		= skapi.AcfNew();
    ItemAcf.acoWearer	= skapi.acoChar;
    ItemAcf.olc		= olcEquipped;

    // Get the aco collection
    var ItemList = ItemAcf.CoacoGet();
    if (BuffDebug2) BuffLog.OutD("Buff_ChkEquip: Type [" + Type + "] Items found [" + ItemList.Count + "]");

    // Now for each item we are wearing of this type, see if it is on the current equipment list.
    // If so then mark it as active, if not then add it to list.
    for (var x=0; x<ItemList.Count; x++) {
 	var EquipmentAco = ItemList.Item(x);
	if (BuffDebug)
	    BuffLog.OutD("Buff_ChkEquip: Checking [" + EquipmentAco.szName + "/" +Util.Hex(EquipmentAco.oid) + "]...");

	for (var Class=equipClassArmor; Class<=equipClassWeapon; Class++) {
	    if (equipEqmFilter[Class] & EquipmentAco.eqm) {
		break;
	    } // endif
	} // endfor

	if (Class == equipClassUnknown) {
	    if (BuffDebug) BuffLog.OutD("Buff_ChkEquip: Item class unknown, skipping..."); 
	    continue;
	} // endif

	// If this item is covenant armor, then skip it since it can't be buffed
	if (EquipmentAco.szName.match(/Covenant/)) {
	    if (BuffDebug) BuffLog.OutD("Buff_ChkEquip: Covenant armor detected, skipped...");
	    continue;
	} // endif

	// See if this piece already on the temporary or permanent equipment lists. First check the perm list.
	for (var y=0; y<this.PermEquipCnt; y++)
	    if (this.PermEquip[y].Oid == EquipmentAco.oid) break;

	// If already on list then...
	if (y<this.PermEquipCnt) {
	    // Mark as active and continue with next item.
	    this.PermEquip[y].Active = true;
	    if (EquipmentAco.eqm & eqmFocusWeapon) ReturnValue = this.PermEquip[y].RendType;
	    if (BuffDebug) BuffLog.OutD("Buff_ChkEquip: Item already on perm equipment list, marked as active.");
	    continue;
	} // endif

	// Now check the temporary list
	for (var y=0; y<this.TempEquipCnt; y++)
	    if (this.TempEquip[y].Oid == EquipmentAco.oid) break;

	// If already on list then...
	if (y<this.TempEquipCnt) {
	    // Mark as active and continue with next item.
	    this.TempEquip[y].Active = true;
	    if (EquipmentAco.eqm & eqmFocusWeapon) ReturnValue = this.TempEquip[y].RendType;
	    if (BuffDebug) BuffLog.OutD("Buff_ChkEquip: Item already on temp equipment list, marked as active.");
	    continue;
	} // endif

	// Item not on list, add it to equipment list as a temporary item, ie user may unequip it at any
	// time which will cause buffs to no longer be cast on it
	var NewIndex = this.TempEquipCnt;
	this.TempEquipCnt++;
	this.TempEquip[NewIndex] = new BuffTempEquipRec(EquipmentAco, Class);
	
	// If this item is a wand, then assess it to see if it has render properties
	if (EquipmentAco.eqm & eqmFocusWeapon) {
	    BuffLog.OutD("Checking wand " + EquipmentAco.szName + " for rending imbue...", ToChat);
	    for (z=0; z<4; z++) {
		var Success = Assess.AssessItem(EquipmentAco);
		if (Success) break;
		if (z<4) BuffLog.OutD("Can't assess item, retrying...", ToChat);
	    } // endfor

	    if (Success) {
		if ((ItemIwi != undefined) && (ItemIwi != null)) {
		    // See if render imbue on wand
		    if (ItemIwi.fSlashRending)		this.TempEquip[NewIndex].RendType = damBlade;
		    if (ItemIwi.fPierceRending)		this.TempEquip[NewIndex].RendType = damPierce;
		    if (ItemIwi.fBludgeonRending)	this.TempEquip[NewIndex].RendType = damBludgeon;
		    if (ItemIwi.fAcidRending)		this.TempEquip[NewIndex].RendType = damAcid;
		    if (ItemIwi.fColdRending)		this.TempEquip[NewIndex].RendType = damCold;
		    if (ItemIwi.fFireRending)		this.TempEquip[NewIndex].RendType = damFire;
		    if (ItemIwi.fLightningRending)	this.TempEquip[NewIndex].RendType = damElectric;
		    ReturnValue = this.TempEquip[NewIndex].RendType;
		    if (ReturnValue > 0) {
			BuffLog.OutD(damName[ReturnValue] + " imbue found on wand.", ToChat);
		    } else {
			BuffLog.OutD("No rending imbue on wand.", ToChat);
		    } // endif
		} else {
		    BuffLog.OutD("No rending imbue on wand.", ToChat);
		} // endif
	    } else {
		BuffLog.OutD("Failed to assess item " + EquipmentAco.szName +
			     ", unable to perform rending check!", ToChat);
	    } // endif
	} // endif
    } // end for

    if (BuffDebug) BuffLog.OutD("Buff_CheckEquip: End");
    return ReturnValue;
} // end func

// Cast one expired buff.  If an expired buff was found then the routine returns 0,
// otherwise it returns the time when the next buff is due
function Buff_Cast(logFlag, spellDurAugGems) {
    if ((logFlag == undefined) || (logFlag == null)) logFlag = false;
    if ((spellDurAugGems == undefined) || (spellDurAugGems == null)) spellDurAugGems = 0;
    if (BuffDebug) BuffLog.OutD("Buff_Cast: Start, logFlag [" + logFlag + "] spellDurAugGems [" +
    				spellDurAugGems + "]");

    var Bonus = 1 + (spellDurAugGems*0.20);
    if (BuffDebug) BuffLog.OutD("Buff_Cast: Bonus [" + Bonus + "]");

    var L_InfoIdx   = -1;	// Index of current leader
    var L_Prio	    = -1;	// Current buff who has highest priority
    var L_EquipIdxP = -1;	// If an item spell, offset on perm equipment list
    var L_EquipIdxT = -1;	// If an item spell, offset on temp equipment list
    var PaperDoll   = false;	// Set to true if the leader's item spell has the paperdoll option set
    var TargetAco   = null;	// Target for item spell
    var ItemSpell   = false;	// If buff found is an item spell then this is set true

    // Reset next time a buff will expire to 0 (which means unknown).
    this.NextExp = 0;

    // First go through the creature and life spells looking for highest priority buff that is expired.
    for (var x=0; x<=this.Range[buffTypeLife].End; x++) {
	with (this.Info[x]) {
	    // if enabled then...
	    if (Enabled) {
	    	if (BuffDebug) BuffLog.OutD("Buff_Cast: Checking [" + Name + "] Type [" + Type + "]");
		if ((Priority > L_Prio) && ((new Date()-LastCast) > buffHoldDown) &&
		    (Force || (EstExp < new Date()))) {
		    L_InfoIdx = x;
		    L_Prio    = Priority;
		    if (BuffDebug)
			BuffLog.OutD("Buff_Cast: New leader: L_InfoIdx [" + L_InfoIdx +  "] L_Prio [" + L_Prio + "]");
		} // endif
		if ((this.NextExp == 0) || (EstExp < this.NextExp)) this.NextExp = EstExp;
	    } // endif
	} // end with
    } // end for
    
    // Check the permanent equipment which is defined by user and loaded from the equipment file.  Permanent 
    // equipment is always buffed if an expired spell is found whether it is being worn or in inventory.
    for (var x=0; x<this.PermEquipCnt; x++) {
	with (this.PermEquip[x]) {
	    if (BuffDebug) BuffLog.OutD("Buff_Cast: Checking Perm Equip [" + Name + "] Eqm [" + Util.Hex(Aco.eqm) + 
					"] Active [" + Active + "] Exists [" + Exists + "]...");

	    // First see if item still exists in inventory
	    if (!Exists) {
		 if (BuffDebug) BuffLog.OutD("Buff_Cast: Item no longer in inventory, skipping...");
		 continue;
	    } // endif

	    // Now go through each possible item spell and see if it is enabled and expired.  Note that
	    // while temporary items use the priority and level from the spell info record, permanently
	    // defined equipment have individually set spell info
	    var ItemStart = this.Range[buffTypeItem].Begin;
	    for (var y=0; y<this.ItemCnt; y++) {
		var Offset = ItemStart+y;
		if (BuffDebug) BuffLog.OutD("Buff_Cast: Checking buff [" + this.Info[Offset].Name + "]...");

		// If buff not enabled, then skip
		if (!Stats[y].Enabled) {
		    if (BuffDebug) BuffLog.OutD("Buff_Cast: Buff disabled, skipping...");
		    continue;
		} // endif

		// If current leaders priority is greater or equal than this buffs prio then skip
		if (L_Prio >= Stats[y].Priority) {
		    if (BuffDebug) BuffLog.OutD("Buff_Cast: Current leader prio >= this buffs prio, skipping...");
		    continue;
		} // endif

		if ((this.NextExp == 0) || (Stats[y].EstExp < this.NextExp)) this.NextExp = Stats[y].EstExp;

		// If force flag is not set, and spell is not expired, then skip
		if (!Stats[y].Force && (Stats[y].EstExp > new Date())) {
		    if (BuffDebug) BuffLog.OutD("Buff_Cast: Spell not expired, skipping...");
		    continue;
		} // endif
		    	
		// If buff has been cast in the last few seconds, then skip
		if ((new Date()-Stats[y].LastCast) < buffHoldDown) {
		    if (BuffDebug) BuffLog.OutD("Buff_Cast: Holddown not expired, skipping...");
		    continue;
		} // endif
		    	 	
		// We have a new leader
		L_InfoIdx   = Offset;
		L_Prio	    = Stats[y].Priority;
		L_EquipIdxP = x;
		ItemSpell   = true;
		TargetAco   = this.PermEquip[L_EquipIdxP].Aco;
		if (BuffDebug)
		    BuffLog.OutD("Buff_Cast: New leader: L_InfoIdx [" + L_InfoIdx + "] L_Prio [" + L_Prio +
				 "] L_EquipIdxP [" +  L_EquipIdxP + "] PaperDoll [" + PaperDoll + "]");
	    } // endfor
	} // end with
    } // end for
    
    // Now go through the temporary equipment list looking for an expired spell.  Temporary equipment definitions
    // are not read in from an external file, but are non-permanent equipment that we are wearing.
    for (var x=0; x<this.TempEquipCnt; x++) {
	with (this.TempEquip[x]) {
	    if (BuffDebug) BuffLog.OutD("Buff_Cast: Checking Temp Equip [" + Name + "] Class [" + Class +
					"] Eqm [" + Util.Hex(Aco.eqm) + "] Active [" + Active + "] ...");

	    // If equipment not active (not being worn) then skip this equipment
	    if (!Active) {
		if (BuffDebug) BuffLog.OutD("Buff_Cast: Equipment not active, skipping...");
		continue;
	    } // endif

	    // Now go through each possible item spell and see if it is enabled and expired.
	    var ItemStart = this.Range[buffTypeItem].Begin;
	    for (var y=0; y<this.ItemCnt; y++) {
		var Offset = ItemStart+y;
		if (BuffDebug) BuffLog.OutD("Buff_Cast: Checking buff [" + this.Info[Offset].Name + "]...");

		// If buff not enabled, then skip
		if (!this.Info[Offset].Enabled) {
		    if (BuffDebug) BuffLog.OutD("Buff_Cast: Buff disabled, skipping...");
		    continue;
		} // endif

		// If current leaders priority is greater than this buffs prio then skip
		if (L_Prio > this.Info[Offset].Priority) {
		    if (BuffDebug) BuffLog.OutD("Buff_Cast: Current leader prio > this buffs prio, skipping...");
		    continue;
		} // endif

		// See if the equipment's eqm matches the spell's eqm, or the spells paper doll option is set
		// and this equipment is armor.
		if (((this.Info[Offset].Eqm & eqmPaperDoll)>0) && (Class == equipClassArmor)) {
		    // Paper doll option set, check against all armor
		    var TestEqm = equipEqmFilter[equipClassArmor];
		    var Doll = true;
		} else {
		    var TestEqm = this.Info[Offset].Eqm;
		    var Doll = false;
		} // endif
		if (BuffDebug) BuffLog.OutD("Buff_Cast: Doll [" + Doll + "] TestEqm [" + Util.Hex(TestEqm) + "]");

		// See if the EQMs overlap. If not then skip
		if (!(TestEqm & Aco.eqm)) {
		    if (BuffDebug) BuffLog.OutD("Buff_Cast: EQMs do not match, skipping...");
		    continue;
		} // endif

		// Keep track of the earliest time a buff will expire.
		if ((this.NextExp == 0) || (Stats[y].EstExp < this.NextExp)) this.NextExp = Stats[y].EstExp;

		// If force flag is not set, and spell is not expired, then skip
		if (!Stats[y].Force && (Stats[y].EstExp > new Date())) {
		    if (BuffDebug) BuffLog.OutD("Buff_Cast: Spell not expired, skipping...");
		    continue;
		} // endif
		    	
		// If buff has been cast in the last few seconds, then skip
		if ((new Date()-Stats[y].LastCast) < buffHoldDown) {
		    if (BuffDebug) BuffLog.OutD("Buff_Cast: Holddown not expired, skipping...");
		    continue;
		} // endif
		    	 	
		// We have a new leader
		L_InfoIdx   = Offset;
		L_Prio	    = this.Info[Offset].Priority;
		L_EquipIdxT = x;
		L_EquipIdxP = -1;	// Cancel any previous permanent equipment leader
		ItemSpell   = true;
		PaperDoll = Doll;
		if (PaperDoll) {
		    TargetAco = skapi.acoChar;
		} else {
		    TargetAco = this.TempEquip[L_EquipIdxT].Aco;
		} // endif
		if (BuffDebug)
		    BuffLog.OutD("Buff_Cast: New leader: L_InfoIdx [" + L_InfoIdx + "] L_Prio [" + L_Prio +
				 "] L_EquipIdxT [" +  L_EquipIdxT + "] PaperDoll [" + PaperDoll + "]");
	    } // endfor
	} // end with
    } // end for

    // Log the final result of the buff search
    if (BuffDebug) BuffLog.OutD("Buff_Cast: Final, L_InfoIdx [" + L_InfoIdx + "] L_Prio [" + L_Prio +
    				"] L_EquipIdxP [" +  L_EquipIdxP + "] L_EquipIdxT [" +  L_EquipIdxT + "]");

    // If no expired buffs found
    if (L_InfoIdx < 0) {
	if (BuffDebug2) BuffLog.OutD("Buff_Cast: End, no expired buffs found");
	return false;
    } // endif

    this.NextExp = 0;

    // If we make it here then buff is expired, so cast buff
    if (logFlag||BuffDebug) {
        var OutLine = "Buff_Cast: Found expired buff [" + this.Info[L_InfoIdx].SpellObj.szName + "] casting";
        if (ItemSpell) {
            OutLine += " on [" + TargetAco.szName + "] ...";
        } else {
            OutLine += "...";
        } // endif
	BuffLog.OutD(OutLine);
    } // endif
    
    // Cast the spell.  The fourth parameter in the cast function indicates
    // whether the spell is a slowcast spell or not.  In this case item buff
    // spells are slow, and creature and life buff spells are fast.  This
    // matches the ItemSpell flag so we use this as the fourth parameter.
    Cast.Cast(this.Info[L_InfoIdx].SpellObj, TargetAco, true, ItemSpell);

    CurTime = new Date();

    if (CastCode != castCodeSuccess) {
	// Spell failed
	if (logFlag||BuffDebug)
            BuffLog.OutD("Buff_Cast: Buff cast failed, CastFail [" + CastFail + "/" +
        		  castFailText[CastFail] + "] CastFail2 [" + CastFail2 + "] Arc [" + CastArc + "]");
        // If cast failed because spell is unknown, then disabled buff
        if ((CastFail == castFailUnknown) && (CastArc == 1022)) {
            if (logFlag||BuffDebug) BuffLog.OutD("Buff_Cast: Buff unknown, buff disabled.");
            this.Info[L_InfoIdx].Enabled = false;
            if (ItemSpell) {
		var Offset = L_InfoIdx - this.Range[buffTypeItem].Begin;
		for (var x=0; x<this.PermEquipCnt; x++) this.PermEquip[x].Stats[Offset].Enabled = false;
	    } // endif
	} // endif
	if (BuffDebug) BuffLog.OutD("Buff_Cast: End, Fail");
	return true;  // Spell attempt was made so return true
    } // endif
  
    // If we make it here, then the spell cast was successfull.  We need to process item spells different
    // then creature and life spells.
    if (logFlag||BuffDebug) BuffLog.OutD("Buff_Cast: Successfull cast [" + this.Info[L_InfoIdx].Name + "]");
    
    if (!ItemSpell) {
	// Life or creature spell
	with (this.Info[L_InfoIdx]) {
	    var OldExp = EstExp;
	    Force = false;
	    Count++;
	    LastCast = new Date();
	    EstExp = CurTime.getTime() + (SpellObj.csecDuration * Bonus * 1000)
	    if (Randomize && (Count == 1)) {
	    	var RandomTime = (SpellObj.csecDuration * Bonus * 1000) * Math.random();
	    	EstExp -= Math.max(RandomTime, Safety);
	    	if (BuffDebug) BuffLog.OutD ("Buff_Cast: Random time [" + RandomTime + "]");
	    } else {
		EstExp -= Safety;
	    } // endif
	    if (BuffDebug) BuffLog.OutD("Buff_Cast: Prev Exp [" + new Date(OldExp) + "] New [" + new Date(EstExp) + "]");
	} // end with
	if (BuffDebug) BuffLog.OutD("Buff_Cast: End, Life/Cre, Success");
	return true;
    } // end if

    // Item spell. First figure out spell offset in stats list
    var SplOff = L_InfoIdx-this.Range[buffTypeItem].Begin;
    // If this is a permanent item, then only have to update one piece since paper doll option not
    // available for permanent items (ie items defined by user in config file).
    if (L_EquipIdxP >= 0) {
	// Permanent item
	if (BuffDebug) BuffLog.OutD("Buff_Cast: Perm item, setting new exp time...");
	with (this.PermEquip[L_EquipIdxP].Stats[SplOff]) {
	    var OldExp = EstExp;
	    Count++;
	    LastCast = new Date();
	    Force = false;
	    EstExp = CurTime.getTime() + (this.Info[L_InfoIdx].SpellObj.csecDuration*Bonus*1000);
	    if (Randomize && (Count==1)) {
		var RandomTime = Math.round((this.Info[L_InfoIdx].SpellObj.csecDuration*Bonus*1000)*Math.random());
		EstExp -= Math.max(RandomTime, Safety);
        	if (BuffDebug) BuffLog.OutD("Buff_Cast: Random time [" + RandomTime + "] Spell Duration [" +
        			    this.Info[L_InfoIdx].SpellObj.csecDuration + "] Bonus [" + Bonus + "]");
	    } else {
		EstExp -= Safety;
	    } // endif
	    if (this.LogBuffs) this.LogItemBuff(L_InfoIdx, Level, EstExp, TargetAco);
	    if (BuffDebug)
		BuffLog.OutD("Buff_Cast: Prev Expire [" + new Date(OldExp) + "] New [" + new Date(EstExp) + "]");
	} // end with
	if (BuffDebug) BuffLog.OutD("Buff_Cast: End, Perm Item, Success");
	return true;
    } // end if

    // If we make it here then we cast an item spell on temporary equipment.  If paper doll option
    // not set then we just update the one piece.  If it is, then we must search through all
    // temporary and permanent armor to see what pieces we are wearing and update the expiration
    // time for each one.
    if (!PaperDoll) {
	// We only cast on one piece so only need to update one spell.
	if (BuffDebug) BuffLog.OutD("Buff_Cast: Temp item (no paperdoll), setting new exp time...");
	with (this.TempEquip[L_EquipIdxT].Stats[SplOff]) {
	    var OldExp = EstExp;
	    Count++;
	    LastCast = new Date();
	    Force = false;
	    EstExp = CurTime.getTime() + (this.Info[L_InfoIdx].SpellObj.csecDuration*Bonus*1000);
	    if ((this.Info[L_InfoIdx].Randomize) && (Count==1)) {
		var RandomTime = Math.round((this.Info[L_InfoIdx].SpellObj.csecDuration*Bonus*1000)*Math.random());
		EstExp -= Math.max(RandomTime, this.Info[L_InfoIdx].Safety);
		if (BuffDebug) BuffLog.OutD("Buff_Cast: Random time [" + RandomTime + "] Spell Duration [" +
				    this.Info[L_InfoIdx].SpellObj.csecDuration + "] Bonus [" + Bonus + "]");
	    } else {
		EstExp -= this.Info[L_InfoIdx].Safety;
	    } // endif
	    if (this.LogBuffs) this.LogItemBuff(L_InfoIdx, this.Info[L_InfoIdx].Level, EstExp, TargetAco);
	    if (BuffDebug)
		BuffLog.OutD("Buff_Cast: Prev Expire [" + new Date(OldExp) + "] New [" + new Date(EstExp) + "]");
	} // end with
	if (BuffDebug) BuffLog.OutD("Buff_Cast: End, Temp Item (no paperdoll), Success");
	return true;
    } // endif

    // If we make it here, then the paper doll eqm option is set and this is a temporary armor item.
    // We need to update the expiration time for all active (equipped) armor equipment on both the
    // permanent and temporary equipment lists.  First calculate a random spell expiration time so
    // that all equipment we are wearing gets the same expiration time
    var RandomTime = Math.round((this.Info[L_InfoIdx].SpellObj.csecDuration*Bonus*1000)*Math.random());
    if (BuffDebug) BuffLog.OutD("Buff_Cast: Paper doll option set, updating all permanent items first.");
    for (var x=0; x<this.PermEquipCnt; x++) {
	with (this.PermEquip[x]) {
	    if (BuffDebug) BuffLog.OutD("Buff_Cast: Perm item (doll), checking [" + Name + "]...");
	    if (!Active) {
	    	if (BuffDebug) BuffLog.OutD("Buff_Cast: Equipment not active, skipping");
		continue;
	    } // endif

	    // Ignore this equipment if it is not armor
	    if (Class != equipClassArmor) {
	    	if (BuffDebug) BuffLog.OutD("Buff_Cast: Equip not armor, skipping");
		continue;
	    } // endif

	    var OldExp = Stats[SplOff].EstExp;
	    Stats[SplOff].Count++;
	    Stats[SplOff].LastCast = new Date();
	    Stats[SplOff].Force = false;
	    Stats[SplOff].EstExp = CurTime.getTime() + (this.Info[L_InfoIdx].SpellObj.csecDuration*Bonus*1000);
	    if (Stats[SplOff].Randomize && (Stats[SplOff].Count==1)) {
		
		Stats[SplOff].EstExp -= Math.max(RandomTime, Stats[SplOff].Safety);
        	if (BuffDebug) BuffLog.OutD("Buff_Cast: Random time [" + RandomTime + "] Spell Duration [" +
        			    this.Info[L_InfoIdx].SpellObj.csecDuration + "] Bonus [" + Bonus + "]");
	    } else {
		Stats[SplOff].EstExp -= Stats[SplOff].Safety;
	    } // endif
	    if (this.LogBuffs) this.LogItemBuff(L_InfoIdx, Stats[SplOff].Level, Stats[SplOff].EstExp, Aco);
	    if (BuffDebug) BuffLog.OutD("Buff_Cast: Prev Expire [" + new Date(OldExp) +
	    				"] New [" + new Date(Stats[SplOff].EstExp) + "]");
	} // end with
    } // end for

    if (BuffDebug) BuffLog.OutD("Buff_Cast: Paper doll option set, now updating all temporary items.");

    for (var x=0; x<this.TempEquipCnt; x++) {
	with (this.TempEquip[x]) {
	    if (BuffDebug) BuffLog.OutD("Buff_Cast: Temp item (doll), checking [" + Name + "]...");
	    if (!Active) {
	    	if (BuffDebug) BuffLog.OutD("Buff_Cast: Equipment not active, skipping");
		continue;
	    } // endif

	    // Ignore this equipment if it is not armor
	    if (Class != equipClassArmor) {
	    	if (BuffDebug) BuffLog.OutD("Buff_Cast: Equip not armor, skipping");
		continue;
	    } // endif

	    var OldExp = Stats[SplOff].EstExp;
	    Stats[SplOff].Count++;
	    Stats[SplOff].LastCast = new Date();
	    Stats[SplOff].Force = false;
	    Stats[SplOff].EstExp = CurTime.getTime() + (this.Info[L_InfoIdx].SpellObj.csecDuration*Bonus*1000);
	    if (this.Info[L_InfoIdx].Randomize && (Stats[SplOff].Count==1)) {
		Stats[SplOff].EstExp -= Math.max(RandomTime, this.Info[L_InfoIdx].Safety);
        	if (BuffDebug) BuffLog.OutD("Buff_Cast: Random time [" + RandomTime + "] Spell Duration [" +
        				 this.Info[L_InfoIdx].SpellObj.csecDuration + "] Bonus [" + Bonus + "]");
	    } else {
		Stats[SplOff].EstExp -= this.Info[L_InfoIdx].Safety;
	    } // endif
	    if (this.LogBuffs) this.LogItemBuff(L_InfoIdx, this.Info[L_InfoIdx].Level, Stats[SplOff].EstExp, Aco);
	    if (BuffDebug) BuffLog.OutD("Buff_Cast: Prev Expire [" + new Date(OldExp) +
	    				"] New [" + new Date(Stats[SplOff].EstExp) + "]");
	} // end with
    } // end for

    if (BuffDebug) BuffLog.OutD("Buff_Cast: End (Doll) Success");
    return true;  // Spell attempt was made so return true
} // end function

function Buff_ResetCreLife() {
   if (BuffDebug) BuffLog.OutD("Buff_ResetCreLife: Begin");

    var ResetCount = 0;
    // Set expire and cast count time to 0 for all creature and life spells
    for (var x=this.Range[buffTypeCre].Begin; x<=this.Range[buffTypeLife].End; x++) {
    	this.Info[x].Count = 0;
    	this.Info[x].EstExp = 0;
    	if (this.Info[x].Enabled) ResetCount++;
    } // endfor

    if (BuffDebug) BuffLog.OutD("Buff_ResetCreLife: End, ResetCount [" + ResetCount + "]");
    return ResetCount;
} // end function


function Buff_ForceAll() {
    if (BuffDebug) BuffLog.OutD("Buff_ForceAll: Begin");

    var ForceCount = 0;
    var ItemBegin = this.Range[buffTypeItem].Begin;
    var ItemEnd   = this.Range[buffTypeItem].End;

    // Set force flag on all spells.  This will cause the immediate casting of all buffs and
    // retrigger the randomization feature. First do creature/life spells
    var ForceCount = 0;
    for (var x=this.Range[buffTypeCre].Begin; x<=this.Range[buffTypeLife].End; x++) {
	if ((this.Info[x].Enabled) && (((new Date()-this.Info[x].LastCast) > this.ForceIgnoreInt))) {
	    if (BuffDebug) BuffLog.OutD("Buff_ForceAll: Forcing spell group [" + this.Info[x].Name + "/" + x + "]");
	    this.Info[x].Force = true;
	    this.Info[x].Count = 0;
	    ForceCount++;
	} // endif
    } // endfor

    // Next force spells on temporary equipment
    for (var x=0; x<this.TempEquipCnt; x++){
    	with (this.TempEquip[x]) {
	    if (BuffDebug) BuffLog.OutD("Buff_ForceUncastItem: Forcing spells on temp equipment [" + Name +
	    				"/" + Util.Hex(Oid) + "]");
	    for (var y=ItemBegin; y<=ItemEnd; y++) {
	    	if ((this.Info[y].Enabled) &&
	    	    (this.Info[y].Eqm & Aco.eqm) &&
	    	    (((new Date()-Stats[y-ItemBegin].LastCast) > this.ForceIgnoreInt))) {
		    Stats[y-ItemBegin].Force = true;
		    ForceCount++;
		    if (BuffDebug) BuffLog.OutD("Buff_ForceUncastItem: Forcing index [" + y + "]");
		} // endif
	    } // endfor
	} // end with
    } // endfor

    // Lastly do permanent equipment
    for (var x=0; x<this.PermEquipCnt; x++){
    	with (this.PermEquip[x]) {
	    if (BuffDebug) BuffLog.OutD("Buff_ForceUncastItem: Forcing spells on perm equipment [" + Name +
	    				"/" + Util.Hex(Oid) + "]");
	    for (var y=ItemBegin; y<=ItemEnd; y++) {
	    	if ((Stats[y-ItemBegin].Enabled) && 
		    (((new Date()-Stats[y-ItemBegin].LastCast) > this.ForceIgnoreInt))) {
		    Stats[y-ItemBegin].Force = true;
		    ForceCount++;
		    if (BuffDebug) BuffLog.OutD("Buff_ForceUncastItem: Forcing index [" + y + "]");
		} // endif
	    } // endfor
	} // end with
    } // endfor

    if (BuffDebug) BuffLog.OutD("Buff_ForceAll: End, ForceCount [" + ForceCount + "]");
    return ForceCount;
} // end function


// Set force flag to true for all item buffs that haven't been cast yet
function Buff_ForceUncastItem() {
    if (BuffDebug) BuffLog.OutD("Buff_ForceUncastItem: Begin");

    var ForceCount = 0;
    var ItemBegin = this.Range[buffTypeItem].Begin;
    var ItemEnd   = this.Range[buffTypeItem].End;
    
    // First force spells on temporary equipment
    for (var x=0; x<this.TempEquipCnt; x++){
    	with (this.TempEquip[x]) {
	    if (BuffDebug) BuffLog.OutD("Buff_ForceUncastItem: Forcing spells on temp equipment [" + Name +
	    				"/" + Util.Hex(Oid) + "]");
	    for (var y=ItemBegin; y<=ItemEnd; y++) {
	    	if ((this.Info[y].Enabled) &&
	    	    (this.Info[y].Eqm & Aco.eqm) &&
	    	    (Stats[y-ItemBegin].Count == 0)) {
		    Stats[y-ItemBegin].Force = true;
		    ForceCount++;
		    if (BuffDebug) BuffLog.OutD("Buff_ForceUncastItem: Forcing index [" + y + "]");
		} // endif
	    } // endfor
	} // end with
    } // endfor

    // Now do permanent equipment
    for (var x=0; x<this.PermEquipCnt; x++){
    	with (this.PermEquip[x]) {
	    if (BuffDebug) BuffLog.OutD("Buff_ForceUncastItem: Forcing spells on perm equipment [" + Name +
	    				"/" + Util.Hex(Oid) + "]");
	    for (var y=ItemBegin; y<=ItemEnd; y++) {
	    	if ((Stats[y-ItemBegin].Enabled) &&
	    	    (Stats[y-ItemBegin].Count == 0)) {
		    Stats[y-ItemBegin].Force = true;
		    ForceCount++;
		    if (BuffDebug) BuffLog.OutD("Buff_ForceUncastItem: Forcing index [" + y + "]");
		} // endif
	    } // endfor
	} // end with
    } // endfor

    if (BuffDebug) BuffLog.OutD("Buff_ForceUncastItem: End, ForceCount [" + ForceCount + "]");
    return ForceCount;
}  // end function

function Buff_Extend(delta) {
    if (BuffDebug) BuffLog.OutD("Buff_Extend: Begin, Delta [" + delta + "]");

   for (var x=0; x<this.EquipCnt; x++) {
    	with (this.Equip[x]) {
	    if (BuffDebug)
		BuffLog.OutD("Buff_Extend: Extending all spells on equipment [" + Name + "/" + Util.Hex(Oid) + "]");
	    for (var y=0; y<buffMaxItemSpells; y++) {
	    	if (Stats[y].EstExp > 0) {
		    Stats[y].EstExp += delta;
		    if (BuffDebug) BuffLog.OutD("Buff_Extend: Extending index [" + x + "]");
	    	} // endif
	    } // endfor
	} // end with
    } // endfor

    if (BuffDebug) BuffLog.OutD("Buff_Extend: End");
} // end func

function Buff_ShowBuffs() {
    BuffLog.OutD("Buff_ShowBuffs: Start");

    // Show info list first

    for (var x=0; x<this.InfoCnt; x++) {
      with (this.Info[x]) {
	if (Enabled) {
	    BuffLog.OutD(" ", true);
	    BuffLog.OutD("Index [" + x + "] Group [" + Name + "] Type [" + Type + "] Level [" + Level + "]", true);
	    BuffLog.OutD("Prio [" + Priority + "] Safety [" + Safety + "] " + "Random [" + Randomize +
	    		 "] Eqm [" + Util.Hex(Eqm) + "]", true);

	    if (Type < 3) {
	    	// Output creature/life spell specific stuff
		var OutLine = "Force [" + Force + "] Count [" + Count + "] EstExp [";
		if (EstExp > 0)
		    OutLine = OutLine + new Date(EstExp);
		else
		    OutLine = OutLine + "0";
		OutLine = OutLine + "]";
		BuffLog.OutD(OutLine, true );
	    } // endif
	} // endif
      } // end with
    } // end for

    // Now dump permanent equipment info
    BuffLog.OutD(" ", true);
    BuffLog.OutD("Permanent equipment count [" + this.PermEquipCnt + "]", true);
    for (var x=0; x<this.PermEquipCnt; x++) {
    	with (this.PermEquip[x]) {
    	    BuffLog.OutD(" ", true);
	    BuffLog.OutD("Name [" + Name + "/" + Util.Hex(Oid) + "] Class [" + Class + "] Exists [" +
	    		 Exists + "] Active [" + Active + "]", true);
	    var Offset = this.Range[buffTypeItem].Begin;			
	    for (var y=0; y<this.ItemCnt; y++) {
	    	if (Stats[y].Enabled) {
		    BuffLog.OutD("Index [" + y + "] Name [" + this.Info[Offset+y].Name + "] Level [" + Stats[y].Level + 
		    		 "] Prio [" + Stats[y].Priority + "] Safety [" + Stats[y].Safety + "]", true);
		    BuffLog.OutD("Randomize [" + Stats[y].Randomize + "] Count [" + Stats[y].Count + "] Exp [" +
		    		  new Date(Stats[y].EstExp) + "] Last Cast [" + new Date(Stats[y].LastCast) +
		    		  "]", true);
		} // endif
	    } // endfor
	} // end with
    } // endfor

    // Lastly dump temporary equipment info
    BuffLog.OutD(" ", true);
    BuffLog.OutD("Temporary equipment count [" + this.TempEquipCnt + "]", true);
    for (var x=0; x<this.TempEquipCnt; x++) {
    	with (this.TempEquip[x]) {
    	    BuffLog.OutD(" ", true);
	    BuffLog.OutD("Name [" + Name + "/" + Util.Hex(Oid) + "] Class [" + Class + "] Active [" + Active + "]", true);
	    var Offset = this.Range[buffTypeItem].Begin;			
	    for (var y=0; y<this.ItemCnt; y++) {
	    	if ((this.Info[Offset+y].Enabled) && (Aco.eqm & this.Info[Offset+y].Eqm)) {
		    BuffLog.OutD("Index [" + y + "] Spl [" + this.Info[Offset+y].Name + "] Count [" +
		    		 Stats[y].Count + "] Exp [" + new Date(Stats[y].EstExp) + "] Last [" +
		    		 new Date(Stats[y].LastCast) + "]", true);
		} // endif
	    } // endfor
	} // end with
    } // endfor

    BuffLog.OutD("Buff_ShowBuffs: End");
    return true;
} // end func


function Buff_ResetPermEquip() {
    this.PermEquipCnt = 0;
} // end function


function Buff_LoadEquip(fileNum) {
    // Load buff settings for permanent equipment from a file
    if (BuffDebug) BuffLog.OutD("Buff_LoadEquip: Begin");

    // Build the output filename from the filenumber.  Note filenum 0 is a special filename.
    if ((fileNum == undefined) || (fileNum == null)) fileNum = 0;
    if (fileNum == 0) {
    	var Post = "def" + fileCfgExt;
    } else {
    	var Post = fileNum + fileCfgExt;
    	if (fileNum < 10) Post = "0" + Post;
    } // endif
    this.FileName = CharCfgDir + "\\" + fileEquipCfg + Post;
 
    // See if file exists
    if (BuffDebug) BuffLog.OutD("Buff_LoadEquip: File name [" + this.FileName + "]");
    if(!this.FileSysObj.FileExists(this.FileName)) {
    	if (BuffDebug) BuffLog.OutD("Buff_LoadEquip: File does not exist");
	return -1;
    } // endif

    // Clear out current buffs
    this.ResetPermEquip();
    this.NextExp = 0;

    // Init misc variables
    var LoadCount = 0;

    // Open file
    this.FileSysObj  = new ActiveXObject("Scripting.FileSystemObject");
    this.TextFileObj = this.FileSysObj.OpenTextFile(this.FileName, 1, false);

    var PrevEquipName	= "";
    var PrevValue	= 0;
    var PrevBurden	= 0;

    var EquipList	= new Array();	// Holds equipment indexes for all equipment we found for an item name.
    var EquipListCnt	= 0;		// Could be multiple items with same name, value, burden
 
    while (!(this.TextFileObj.AtEndOfStream)) {
	// now read lines until EOF
	var Line  	= this.TextFileObj.ReadLine();
	Line = Util.Trim(Line)
	if ((Line.substr(0,2) == "//") || (Line.substr(0,1) == "#")) {
	    if (BuffDebug) BuffLog.OutD("Buff_LoadEquip: Comment [" + Line + "]");
	    continue;		// Comment character at beginning of line, skip it
	} // endif
	if (BuffDebug) BuffLog.OutD("Buff_LoadEquip: Line [" + Line + "]");
	var Split 	= Line.split(",");
	var EquipName	= Util.Trim(Split[0]);
	var Value	= parseInt(Split[1]);
	var Burden	= parseInt(Split[2]);
	var BuffName	= Util.Trim(Split[3]);
	var Level	= parseInt(Split[4]);
	var Priority	= parseInt(Split[5]);
	var Safety	= parseInt(Split[6]);
	if (Util.Trim(Split[7]) == "true")
	    var Randomize = true;
	else
	    var Randomize = false;
	    
	if (BuffDebug) {
	    BuffLog.OutD("Buff_LoadEquip: EquipName [" + EquipName + "] Value [" + Value + "] Burden [" + Burden +
	    		 "] BuffName [" + BuffName + "]");	 
	    BuffLog.OutD("Buff_LoadEquip:  Level [" + Level + "] Prio [" + Priority + "] Safety [" +
	    		 Safety + "] Randomize [" + Randomize + "]");
	} // endif 

	// If this is not the same piece of equipment as last time, then we need to look up this
	// item in inventory.  Could be multiple matches.
	if ( (EquipName != PrevEquipName) || (Value != PrevValue) || (Burden != PrevBurden)) {
	    PrevEquipName 	= EquipName;
	    PrevValue	  	= Value;
	    PrevBurden	  	= Burden;

	    var EquipList	= new Array();	// Holds equipment indexes for all equipment we found for an item name.
	    var EquipListCnt	= 0;		// Could be multiple items with same name, value, burden

	    // Find equipment in inventory
	    var acf = skapi.AcfNew();
	    acf.olc    = olcInventory;
	    acf.szName = EquipName;
	    if (Value>0) {
		acf.cpyMin = Value;
		acf.cpyMax = Value;
	    } // endif
	    if (Burden>0) {
		acf.burdenMin = Burden;
		acf.burdenMax = Burden;
	    } // endif  
	    
	    var InventList = acf.CoacoGet();
	    if (InventList.Count == 0) {
	    	// Unknown item, skipping
	    	BuffLog.OutDM("Buff_LoadEquip", "Can't find equipment [" + EquipName + "] in inventory, skipping", true);
	    	continue;
	    } else {
		if (BuffDebug) BuffLog.OutD("Found [" + InventList.Count + "] in inventory");
	    } // endif

	    // For each item found, add it to the permament equipment list if it is not already on it
	    for (var x=0; x<InventList.Count; x++) {
		// See if this item is already on the perm equipment list
		for (var y=0; y<this.PermEquipCnt; y++)
		    if (InventList.Item(x).oid == this.PermEquip[y].Oid) break;

		if (y == this.PermEquipCnt) {
		    // Item not found, add it to the perm equipment list
		    LoadCount++;	// Increment count of equipment pieces found
		    this.PermEquip[y] = new BuffPermEquipRec(InventList.Item(x));
		    this.PermEquipCnt++;
		    if (InventList.Item(x).eqm & eqmFocusWeapon) {
		    	BuffLog.OutD("Checking perm wand " + InventList.Item(x).szName + " for rending imbue...");
			for (z=0; z<4; z++) {
			    var Success = Assess.AssessItem(InventList.Item(x));
			    if (Success) break;
			    if (z<4) BuffLog.OutD("Can't assess item, retrying...");
	    		} // endfor

			if (Success) {
	    		    if ((ItemIwi != undefined) && (ItemIwi != null)) {
				// See if render imbue on wand
				if (ItemIwi.fSlashRending)	this.PermEquip[y].RendType = damBlade;
				if (ItemIwi.fPierceRending)	this.PermEquip[y].RendType = damPierce;
				if (ItemIwi.fBludgeonRending)	this.PermEquip[y].RendType = damBludgeon;
				if (ItemIwi.fAcidRending)	this.PermEquip[y].RendType = damAcid;
				if (ItemIwi.fColdRending) 	this.PermEquip[y].RendType = damCold;
				if (ItemIwi.fFireRending)	this.PermEquip[y].RendType = damFire;
				if (ItemIwi.fLightningRending)  this.PermEquip[y].RendType = damElectric;
				if (this.PermEquip[y].RendType > 0) {
				    BuffLog.OutD(damName[ReturnValue] + " imbue found on wand.");
				} else {
				    BuffLog.OutD("No rending imbue on wand.");
				} // endif
			    } else {
			    	BuffLog.OutD("No rending imbue on wand.");
			    } // endif
			} else {
			    BuffLog.OutD("Failed to assess perm item " + this.PermEquip[y].szName +
			    		", unable to perform rending check!");
			} // endif
		    } // endif
		    if (BuffDebug) BuffLog.OutD("Buff_LoadEquip: Added equipment [" + InventList.Item(x).szName + "]");
		} else {
		    if (BuffDebug) BuffLog.OutD("Buff_LoadEquip: Equipment already on perm list.");
		} //endif
		EquipList[EquipListCnt] = y;
	    	EquipListCnt++;
	    } // endif
	} // endif
    
        if (BuffDebug) {
	    BuffLog.OutD("Buff_LoadEquip: EquipListCnt [" + EquipListCnt + "]");
	    for (var x=0; x<EquipListCnt; x++)
		 BuffLog.OutD("Buff_LoadEquip: EquipList[" + x + "] Value [" + EquipList[x] + "]");
	} // endif
    
	var BuffIndex = this.Lookup(BuffName);
	if (BuffIndex < 0) {
	    BuffLog.OutD("Unknown buff name [" + BuffName + "] for equipment [" + EquipName + "]", true);
	    continue;
	} // endif
	
	// Now get the spell offset in the item buff range.  If less than 0 then we have a creature
	// or life spell so reject buff.
	Offset = BuffIndex - this.Range[buffTypeItem].Begin;
	if (Offset < 0 ) {
	    BuffLog.OutD("Only item buffs allowed for equipment. Buff [" + BuffName +
	    		 "] invalid on [" + EquipName + "]", true);
	    continue;
	} // endif
	
	if (BuffDebug) BuffLog.OutD("Buff_LoadEquip: Buff index [" + BuffIndex + "] Offset [" + Offset + "]");
	for (var x=0; x<EquipListCnt; x++) {
	    with (this.PermEquip[EquipList[x]]) {
	    	if (BuffDebug) BuffLog.OutD("Buff_LoadEquip: Setting equipment at index [" + EquipList[x] + "]");
		Stats[Offset].Enabled	= true;
		Stats[Offset].Level	= Level;
		Stats[Offset].Priority	= Priority;
		Stats[Offset].Safety	= Safety;
		Stats[Offset].Randomize	= Randomize;
		Stats[Offset].SpellObj	= skapi.SpellInfoFromSpellid(this.Info[BuffIndex].Self[Level]);

		if (!Cast.Know(this.Info[BuffIndex].Self[Level])) {
		    Stats[Offset].Enabled = false;
		    Chat.Screen("You do not know " + this.Info[BuffIndex].Name + " Self " +
				 Stats[Offset].Level + ", buff disabled on [" + EquipName + "]");
		} // endif

	    } // end with
	} // end for
    } // end while
    
    this.TextFileObj.Close();

    if (BuffDebug) BuffLog.OutD("Buff_LoadEquip: End, LoadCount [" + LoadCount + "]");
    return LoadCount;
} // endif


function Buff_LogItemBuff(SpellIndex, Level, ExpTime, TargetAco ) {
    if (BuffDebug)
	BuffLog.OutD("Buff_LogItemBuff: Begin, SpellIndex [" + SpellIndex + "] Target [" + TargetAco.szName + "]");

    // Write: SpellName, Level, ExpTime, Target Oid
    this.LogFileObj.WriteLine(this.Info[SpellIndex].Name + ", " + Level + ", " + ExpTime + ", " + TargetAco.oid );
   
    if (BuffDebug) BuffLog.OutD("Buff_LogItemBuff: End");
} // end function


// Turn item buff logging on or off
function Buff_SetBuffLog(Enabled, LogFile) {
    if (BuffDebug) BuffLog.OutD("Buff_SetBuffLog: Begin");
    if ((LogFile != undefined) && (LogFile != null) && (LogFile != "")) {
	if (BuffDebug) BuffLog.OutD("Buff_SetBuffLog: Log file set to [" + LogFile + "]");
	this.LogFile = LogFile;
    } // endif

    // If item buff logging not enabled previously and we are enabling it then we need to create the item
    // buff log file and populate it with any item spell currently active temp or perm equipment lists.
    if (!this.LogBuffs && Enabled) {
	this.LogFileObj = this.FileSysObj.OpenTextFile(this.LogFile, ForWriting, true);
	this.LogBuffs = true;
	var Offset = this.Range[buffTypeItem].Begin;

	// Write any active item buffs to the log file.  First search through items on perm list.
	for (var x=0; x<this.PermEquipCnt; x++) {
	    with (this.PermEquip[x]) {	
		for (var y=0; y<this.ItemCnt; y++) {
		    if ((Stats[y].Enabled) && (Stats[y].EstExp > new Date())) {
			this.LogItemBuff(Offset+y, Stats[y].Level, Stats[y].EstExp, Aco);
		    } // endif
		} // endfor
	    } // end with
	} // endfor

	// Now go through equipment on the temporary list
	for (var x=0; x<this.TempEquipCnt; x++) {
	    with (this.TempEquip[x]) {		
		for (var y=0; y<this.ItemCnt; y++) {
		    if ((this.Info[Offset+y].Enabled) &&
			(Aco.eqm & this.Info[Offset+y].Eqm) &&
			(Stats[y].EstExp > new Date())) {
		        this.LogItemBuff(Offset+y, this.Info[Offset+y].Level, Stats[y].EstExp, Aco);
		    } // endif
		} // endfor
	    } // end with
	} // endfor
    } // endif
    
    // If item item buff logging previously enabled and we are disabling it,
    // then delete the buff log file if it exists
    if (this.LogBuffs && !Enabled) {
    	this.LogFileObj.Close();
	this.FileSysObj.DeleteFile(this.LogFile);
	this.LogBuffs = false;
    } // endif

    if (BuffDebug) BuffLog.OutD("Buff_SetBuffLog: End");
} // end function


function Buff_ProcItemLog(inputFile) {
    if (BuffDebug) BuffLog.OutD("Buff_ProcItemLog: Begin");
    if (!this.FileSysObj.FileExists(inputFile)) return false;

    var TextFileObj = this.FileSysObj.OpenTextFile(inputFile, ForReading, false);
    
    // Buff name, buff level, Exp Time, Oid
    // 
    // Read records in file until EOF
    while (!(TextFileObj.AtEndOfStream)) {
	var Line = TextFileObj.ReadLine();
	if (BuffDebug) BuffLog.OutD("Buff_ProcItemLog: Processing line [" + Line + "]");
	var Split  = Line.split(",");
	var Name   = Util.Trim(Split[0]);
	var Level  = parseInt(Split[1]);
	var EstExp = parseInt(Split[2]);
	var Oid    = parseInt(Split[3]);

	var BuffIndex = this.Lookup(Name);
	if (BuffIndex < 0) {
	    BuffLog.OutD("ProcItemLog: Unknown buff name [" + Name + "]", true);
	    continue;
	} // endif
	var Offset = BuffIndex - this.Range[buffTypeItem].Begin;

	// See if spell expired
	if (EstExp < new Date()) {
	    if (BuffDebug) BuffLog.OutD("Buff_ProcItemLog: Spell expired");
	    continue;
	} // endif

	// See if this item is on the perm equipment list
	for (var y=0; y<this.PermEquipCnt; y++)
	    if (Oid == this.PermEquip[y].Oid) break;
	if (y < this.PermEquipCnt) {
	    // Item found, update spell expiration time
	    this.PermEquip[y].Stats[Offset].EstExp = EstExp;
	    continue;
	} // endif
	    
	// Next look on temporary equipment list
	for (var y=0; y<this.TempEquipCnt; y++)
	    if (Oid == this.TempEquip[y].Oid) break;
	    
	// If item not found, then add it to temporary list if it is in our inventory,
	// but mark it as inactive ie not being worn
	if (y == this.TempEquipCnt) {
	    var EquipmentAco = skapi.AcoFromOid(Oid);
	    if ((EquipmentAco == undefined) || (EquipmentAco == null)) {
	    	// Item not detectable, skip it
	    	continue;
	    } //endif
	    if (!EquipmentAco.fInInventory) {
		// Item not in inventory, skip it
		continue;
	    } // endif
	    for (var Class=equipClassArmor; Class<=equipClassWeapon; Class++) {
		if (equipEqmFilter[Class] & EquipmentAco.eqm) {
		    break;
		} // endif
	    } // endfor
	    this.TempEquip[y] = new BuffTempEquipRec(EquipmentAco, Class);
	    this.TempEquip[y].Active = false;
	    this.TempEquipCnt++;
	} // endif

	// Update spell expiration time
	this.TempEquip[y].Stats[Offset].EstExp = EstExp;

    } // end while

    // Close file
    TextFileObj.Close();
    
    if (BuffDebug) BuffLog.OutD("Buff_Init: End");
    return true;
} // end function

Buff = new BuffObj(false);