// MD_CASTOBJ.JS
//
// Author:	Poundboy
// Email:	poundboy@megapathdsl.net
// Version:	1.01
//
// Description:
//
// The following routines implement a spell Cast object around the skapi.CastSpell method.  Note this
// object uses a combination of the new skapi.arc property with the old acscript style of checking for
// spell completion.  This object has the following features:
//
// Attempt timout	If the OnChatSpell event is not received within a user settable time (default
//			is 2 seconds), then the cast attempt aborts.  This is useful when the spell
//			target moves out of range or dies just before cast begins but after target
//			is selected.
//
// Completion timeout	If the spell is not cast by a user settable time (default variable by spell
//			level), then assume the spell failed.  This is useful in cases where the
//			OnChatSpell event is received but no spell completion event is ever received.
//			This can happen due to lag, dropped events, or the server side bug where
//			you see the spell words but the spell never happens.
//
// Cast Abort		If in the middle of a spell cast you decide to abort, then calling this
//			method will cause the cast attempt to return immediately with an abort
//			error code.  This is useful when you need to react to a critical event asap,
//			for example you are taking melee damage and need to log or run away, but
//			you are in the middle of a spell cast.
//
// Failure counts	Keeps track of the number of times each spell fails to cast.  A successful
//			spell cast clears the failure count for that spell index.
//
// Requirements:
//
// Requires the following libraries: MD_UTIL.JS, MD_LOGOBJ.JS, MD_SPELLOBJ.JS
//
// Also requires MD_PLAYEROBJ.JS if player combat mode/peace mode checking enabled
//
//
// Object Constructor:
//
// myObj = new CastObj(DebugFlag)
//
// where DebugFlag is true or false and controls whether debug info written to console and optionally log file
//
// Methods:
//
// Cast(SpellObj,target,castmsgreq,slowcast)
//					Casts a spell by skapi spell object on optional target.
//					Castmsgreq indicates whether the spell always returns a 'You cast...' message.
//					Slowcast indicates the spell is a quick self creature or life buff or not
// GetSpellObj(SpellName)		Converts a spells name to its spell object
// Abort()				Abort casting of a spell
// FailClear(SpellId)			Clear a spell's fail count
// FailCount(SpellId)			Return a spell's fail count
// AttemptTimeout(value)		Set attempt timeout value
// CompletionTimout(level,value)	Set completion timeout value for a level
// SetDebug(flag,opm)			Change debug flag/output
//
//
// Modification History
//
// Who	Date	  Version	Description
//
// PB	06/23/02  1.01		Initial version
//
// PB	09/21/02  1.02		Use skapi spell object instead of spell ids for casting
//
// *******************************************
// Global  Variables
//
// Constants
//

// Spell cast status. Current Status of Spell Casting. Possible values:
//
// castState_None		No spell being cast.
// castState_Attempting		Spell cast is being attempted.
// castState_Chatspell		OnChatSpell message received (ex: Maud casts Malar Caril). This
//				means the animation has started and we should expect a completion
//				event in 1-5 seconds (actual time depends on level of spell).
// castState_Done		Spell completed either successfully or failed.  See Spell_Code.
castStateNone		= 1;
castStateAttempting	= 2;
castStateChatSpell	= 3;
castStateDone		= 4;

var castStateText = new Array();
castStateText[castStateNone]	   = "None";
castStateText[castStateAttempting] = "Attempting";
castStateText[castStateChatSpell]  = "ChatSpell";
castStateText[castStateDone]	   = "Done";

// Spell cast result codes
// If CastState = castStateDone, then the following are possible results:
//
// castCodeUnknown		Spell cast status not known yet
// castCodeSuccess		The spell cast was successful
// castCodeFail			The spell cast failed.  Reason is in
//				Spell_FailCode
castCodeUnknown		= 0;
castCodeSuccess		= 1;
castCodeFail		= 2;

var castCodeText = new Array();
castCodeText[castCodeUnknown]	= "Unknown";
castCodeText[castCodeSuccess]	= "Success";
castCodeText[castCodeFail]	= "Fail";

// Spell cast failure codes.  Spell failed for one of the following reasons:
//
// castFailBusy		Unable to cast spell due to character being busy (SpellFailSelf reason=29)
// castFailFizzle	Spell Fizzled
// castFailNoComp	Out of one or more comps needed for spell
// castFailNoMana	Not enough mana to cast spell
// castFailOutRange	Target out of range
// castFailResist	Target resisted spell
// castFailTimeout	No spell status received
// castFailAbort	Spell was aborted by macro
// castFailArc		Couldn't determine spell failure by looking at events, but
//			skapi.arc figured it out somehow.
// castFailTooShort	Sometimes when casting a spell, words appear but spell never casts
//			and event code mark it as a success, so we check for a minimum time
//			for all spell casts
// castFailUnknown	Can't figure out failure reason
castFailUndefined	= 0;
castFailBusy		= 1;
castFailFizzle		= 2;
castFailNoComp		= 3;
castFailNoMana		= 4;
castFailOutRange	= 5;
castFailResist		= 6;
castFailTimeout		= 7;
castFailAbort		= 8;
castFailArc		= 9;
castFailTooShort	= 10;
castFailNoCastMsg	= 11;
castFailUnknown		= 12;

var castFailText = new Array();
castFailText[castFailUndefined]	= "Undefined";
castFailText[castFailBusy]	= "Busy";
castFailText[castFailFizzle]	= "Fizzle";
castFailText[castFailNoComp]	= "NoComp";
castFailText[castFailNoMana]	= "NoMana";
castFailText[castFailOutRange]	= "OutRange";
castFailText[castFailResist]	= "Resist";
castFailText[castFailTimeout]	= "Timeout";
castFailText[castFailAbort]	= "Abort";
castFailText[castFailArc]	= "Arc";
castFailText[castFailTooShort]	= "TooShort";
castFailText[castFailNoCastMsg]	= "NoCastMsg";
castFailText[castFailUnknown]	= "Unknown";

castFailIntMax			= 60*1000;	// If we get consectutive spell failures for this many milliseconds,
						// when we try to move our char a little bit ('unsticks' us in
						// certain cases).
castFailLastMax			= 10*1000;	// If last spell was cast more than this many milliseconds ago,
						// then clear out the consecutive failure start timestamp.

// Global variables.  Can't put most of these in objects since events fire outside of object context
var CastComplete	= false;		// Set to true by OnActionComplete event
var CastDebug		= false;		// Set to true when cast debug info desired
var CastLog		= new LogObj();		// Create log object for comp routines
var Casth		= new CastHandlers();	// Set up handlers variable as a global variable
var CastHandReg		= false;		// Set to true when handlers registered
var CastState		= castStateNone;	// See above description of SpellState constants
var CastCode		= 0;			// See above description of SpellCode constants
var CastFail		= 0;			// See above description of SpellFail constants'
var CastFail2		= 0;			// If fail code unknown, put for reason code in here from the event
var CastArc		= 0;			// Spells skapi.arc result
var CastAbortFlag	= false;		// If set spell cast attempt will be aborted
var CastFailCnt		= 0;			// Number of times spell failure movement reset executed
var CastActive		= false;		// Set to true when spell cast active

// ********************************************************************************************
// Spell object event Functions. Dispatched via skapi.WaitEvent()

// OnActionComplete
function CastHandlers_OnActionComplete() {
    if (CastDebug) CastLog.OutD("Cast_OnActionComplete: Begin" );
    
    if (CastDebug) CastLog.OutD("Cast_OnActionComplete: Setting Castcomplete to true" );
    CastComplete = true;

    if (CastDebug) CastLog.OutD("Cast_OnActionComplete: End" );
} // end func


// OnChatServer
function CastHandlers_OnChatServer(text, code) {
    if (CastDebug) CastLog.OutD("Cast_OnChatServer: Text (" + text + ") Type (" + code + ")");

    // If we are not casting a spell, then just return
    if (CastState == castStateDone) return;

    if (text.match(/resists your spell/)) {
	if (CastDebug) CastLog.OutD("Cast_OnChatServer: target resisted");
	CastState	= castStateDone;
	CastCode	= castCodeFail;
	CastFail	= castFailResist
	return;
    } // endif

    if (text.match(/out of range/)) {
	if (CastDebug) CastLog.OutD("Cast_OnChatServer: Out of range detected" );
	CastState	= castStateDone;
	CastCode	= castCodeFail;
	CastFail	= castFailOutRange;
	return;
    } // endif

    if ((text.match(/You cast/)) ||
        ((text.match(/gain/))&&(text.match(/casting/))) || (text.match(/you drain/))) {
	if (CastDebug)
	    CastLog.OutD("Cast_OnChatServer: Detected cast. Spell Done & Successful");
	CastState 	= castStateDone;
	CastCode	= castCodeSuccess;
	return;
    } // endif

      if (CastDebug) CastLog.OutD("Cast_OnChatServer: End" );

} // end function

// OnChatSpell
//
function CastHandlers_OnChatSpell( acoSender, Text ) {
    if (CastDebug)
	CastLog.OutD("Cast_OnChatSpell: Id [" + Util.Hex(acoSender.oid) + "] Who [" +
			acoSender.szName + "] Text [" + Text + "]");
    if (skapi.szName == acoSender.szName) {
    	 if (CastDebug) CastLog.OutD("Cast_OnChatSpell: Cast_State set to ChatSpell");
    	 CastState = castStateChatSpell;
    } // endif
} // end func

// OnSpellFailSelf
//
function CastHandlers_OnSpellFailSelf(reason) {
	if (CastDebug) CastLog.OutD("Cast_OnSpellFailSelf: Reason (" + reason + ")");
	CastState = castStateDone;
	CastCode  = castCodeFail;
	switch (reason) {
		case 29:	CastFail  = castFailBusy;
			 	break;	// immediate

		case 1024:	CastFail  = castFailNoComp;
				break;	// immediate

		case 1025:	CastFail  = castFailNoMana;
				break;	// immediate

		case 1026:	CastFail  = castFailFizzle;
				break;	// after animation

		default:	CastFail  = castFailUnknown;
				CastFail2 = reason;
				break;
	} // end switch
	if (CastDebug) CastLog.OutD("Cast_OnSpellFailSelf: End");
} // end func

function CastHandlers() {
	this.OnActionComplete	= CastHandlers_OnActionComplete;
	this.OnChatServer	= CastHandlers_OnChatServer;
	this.OnChatSpell	= CastHandlers_OnChatSpell;
	this.OnSpellFailSelf	= CastHandlers_OnSpellFailSelf;
}


function CastObj(debugflg) {
    if ((debugflg==undefined) || (debugflg==null)) debugflg = false;
    CastDebug = debugflg;

    // War/Peace mode check flag.  If set to true then the cast functions will ensure we are
    // in war mode before casting
    this.ModeCheckFlag	= true;

    // Add handlers for events we are interested in for this object.  Only add them if this
    // is the first time this object was created
    if (!(CastHandReg))	{
	skapi.AddHandler(evidOnActionComplete,	Casth );
	skapi.AddHandler(evidOnChatServer,	Casth );
	skapi.AddHandler(evidOnChatSpell,	Casth );
	skapi.AddHandler(evidOnSpellFailSelf,	Casth );
	CastHandReg = true;
    } // endif

    // If we dont get a OnChatSpell within x milliseconds, then assume spell was never cast.
    // This could happen if creature was killed by someone else, left our target range, or we
    // encountered the infamous casting bug where the animation starts but no spell is cast.  Even
    // though we have a timeout for the whole spell cast attempt, that value is much larger
    // than this timeout, so the spell failure is detected much quicker which is important in
    // a melee or combat macro
    this.AttemptTimeout = 4000;

    // If Cast_State = Attempting, this variable holds the time the spell cast was
    // initiated. We use this this variable to determine when we should assume the
    // spell failed or succeeded but we missed the final message.
    this.AttemptTime = new Date();

    // Array for keeping track of spell failures for each spell Id
    this.FailList = new Array();

    // Array of times (ms) for each spell level when spell will be declared aborted.
    // Level is calculated from spell difficulty so thats why there is lvl 8 and 9 entries in the array,
    // ie some spells are more difficult then lvl7s (like self debuff spells).
    this.Timeout = new Array(0, 4000, 5000, 6000, 7000, 9000, 10000, 10000, 10000, 10000);

    // Array of times (ms) for each spell level that casting must be at least this long in length.
    // This catches problem where the spell animation begins but no spell is cast and the server
    // returns a success code almost immediately after the cast begis. Level is calculated from spell
    // difficulty so thats why there is lvl 8 and 9 entries in the array, ie some spells are more
    // difficult then lvl7s (like self debuff spells).  Note that spell durations are much
    // shorter for self buff life and creature spells.
    this.MinTimeFast = new Array(0, 500,   500,  500,  500,  500,  500,  500,  500,  500);
    this.MinTimeSlow = new Array(0, 1000, 2000, 2000, 2500, 2500, 3000, 3000, 3000, 3000);

    // Keep track of the number of consecutive spell failures.  If we get too many, then we move
    // around a bit in case we are 'stuck'/
    this.FailCount = 0;

    // Set when we detect a spell failure and cleared when we have a success.
    // Used to detect when we are getting repeated failures.
    this.FirstFail = 0;

    // Now create an array of all spells in this characters spellbook.  This array is indexed by spellid
    // and the value is set to TRUE if the spell exists in our spellbook.  Note you must call BuildSpellBook
    // to initialize this array because the user may not be logged in when the object is created.
    this.SpellBook = new Array();

    // Public Methods
    this.SetDebug		= Cast_SetDebug;
    this.Cast			= Cast_Cast;		// Casts a spell by spell object
    this.Know			= Cast_Know;		// Checks to see if we know a spell or not
    this.GetSpellObj		= Cast_GetSpellObj;	// Converts spell text to spell object
    this.ModeCheck		= Cast_ModeCheck;	// Sets mode check flag
    this.Abort			= Cast_Abort;		// Abort casting of a spell
    this.FailClear		= Cast_FailClear;	// Clear a spells fail count
    this.FailCount		= Cast_FailCount;	// Return a spells fail count
    this.SetAttTimeout		= Cast_SetAttTimeout;	// Set attempt timeout value
    this.SetCompTimeout		= Cast_SetCompTimout;	// Set completion timeout value for a level
    this.BuildSpellBook		= Cast_BuildSpellBook;	// Build spell book

} // end func CastObj

function Cast_BuildSpellBook() {
    if (CastDebug) CastLog.OutD("Cast_BuildSpellBook: Begin");

    var SpellCnt = skapi.rgspellid(-1,-1);
    if (CastDebug) CastLog.OutD("SpellCnt [" + SpellCnt + "]");

    // Set boolean flag for each spell we have in our spellbook
    for (var Index=0; Index<SpellCnt; Index++) {
    	var SpellId = skapi.rgspellid(-1,Index);
    	if (CastDebug) CastLog.OutD("Index [" + Index + "] SpellId [" + SpellId + "] SpellName [" +
	    				skapi.SzFromSpellid(SpellId) + "]");
    	this.SpellBook[SpellId] = true;
    } //end for
    if (CastDebug) CastLog.OutD("Cast_BuildSpellBook: End");
} // end func

function Cast_SetDebug(debugflag) {
    CastDebug = debugflag;
    return true;
}

// This function could be probably be a lot shorter by just using the OnActionComplete event in
// conjunction with the skapi.arc property but I don't trust them completely (yet), so I first
// try to use the old fashion way that I did with acscript, then use skapi.arc as a backup
// if my way fails.
function Cast_Cast(SpellObj, TargetAco, CastMsgReq, SlowCast) {
    if (TargetAco == undefined) TargetAco = null;
    if ((CastMsgReq == undefined)  || (CastMsgReq == null)) CastMsgReq = false;
    if ((SlowCast   == undefined)  || (SlowCast   == null)) SlowCast   = false;
    
    if (CastDebug) {
    	CastLog.OutD("Cast_Cast: Begin, SpellId [" + SpellObj.spellid + "] SpellName [" +
    		      SpellObj.szName + "] CastMsgReq [" + CastMsgReq + "] SlowCast [" + SlowCast + "]");
    	if (TargetAco == null) {
    	    CastLog.OutD("Cast_Cast: No target specified");
    	} else {
    	    CastLog.OutD("Cast_Cast: Target Name [" + TargetAco.szName + "]");
    	} // endif
    	
    } // endif

    if (this.FirstFail != 0) {
    	// If the past n casts were failures, make sure those casts were recent attempts.
    	// If not then reset the spell failure start timer.
	if ((new Date()-CastEndTime) > castFailLastMax) this.FirstFail = 0;
    } // endif

    // Compute spells level from difficulty
    var Level	= Math.round(SpellObj.diff/50)+1;
    var Name    = SpellObj.szName;
    var Timeout = this.Timeout[Level];		// Get timeout value for spell being cast.
    if (CastDebug)
	CastLog.OutD("Cast_Cast: Level [" + Level + "] Name [" + Name + "] Timeout [" + Timeout + "]");

    // Make sure we are in magic mode if checkmode flag set
    if (this.ModeCheckFlag) Player.EnsureCombatMode(true);

    // Now reset all global variables that track spell cast progress
    CastArc 		= skapi.arc;			// Reset last action code
    CastComplete 	= false;			// Reset action complete flag.
    CastCode 		= castCodeUnknown;		// Reset spell success/fail code
    CastFail		= 0;				// Reason spell failed
    CastFail2		= 0;				// Reason if CastFail = Unknown
    CastAbortFlag	= false;			// Reset spell abort flag
    CastActive		= true;				// Mark spell cast as being in progress

    // Now set Cast_State to indicate spell cast attempt in process and record start time
    CastState = castStateAttempting;	// Set Cast_State to indicate we are attempting a spell.

    var CastStartTime = new Date();	// Save time spell attempt was made so we can
					// determine when to give up (timeout limit reached)

    // Finally cast the dam spell
    if (CastDebug) CastLog.OutD("Cast_Cast: Attempting to cast spell...");
    if (TargetAco == null)
    	skapi.CastSpell(SpellObj.spellid);
    else
        skapi.CastSpell(SpellObj.spellid, TargetAco);

    // Loop until we get event messages that indicate a completion of a spell (success or
    // failure) or we reach the timeout threshold
    if (CastDebug) CastLog.OutD("Cast_Cast: Waiting for events or timeout condition...");
    while ((CastState != castStateDone) && !CastAbortFlag && !CastComplete) {
    	skapi.WaitEvent(10, wemFullTimeout);
	if (CastState != castStateDone) {
	    if ((CastState == castStateAttempting) &&
		((new Date() - CastStartTime) > this.AttemptTimeout)) {
	        if (CastDebug) CastLog.OutD  ("Cast_Cast: Attempt timeout");
	        CastState    = castStateDone;
	        CastCode     = castCodeFail;
	        CastFail     = castFailTimeout;
	        CastComplete = true;
	        break;
	    } // endif

	    if ((new Date() - CastStartTime) > Timeout) {
	        if (CastDebug) CastLog.OutD  ("Cast_Cast: Timeout");
	        CastState    = castStateDone;
	        CastCode     = castCodeFail;
	        CastFail     = castFailTimeout;
		CastComplete = true;
		break;
	     } // endif
        } // endif
    } // end while

    // Mark spell cast as being pretty much finished.  Used by war
    // evasion routines to determine when to slidecast.
    CastActive = false;

    if (CastAbortFlag) {
        if (CastDebug) CastLog.OutD("Cast_Cast: Abort detected");
    	CastState = castStateDone;
	CastCode  = castCodeFail;
	CastFail  = castFailAbort;
	return;
    } // endif

    // Now wait for animation to complete or spell timeout to be reached.
    // Easy to do this now with the new OnActionComplete event.
    if (CastDebug) CastLog.OutD("Cast_Cast: Waiting for OnActionComplete event...");
    while (!CastComplete && ((new Date()-CastStartTime) < Timeout) && !CastAbortFlag) {
	skapi.WaitEvent(10, wemFullTimeout);  // Process all events in queue
    } // end while

    var CastEndTime = new Date();

    if (CastAbortFlag) {
        if (CastDebug) CastLog.OutD("Cast_Cast: Abort detected");
    	CastState = castStateDone;
	CastCode  = castCodeFail;
	CastFail  = castFailAbort;
	return;
    } // endif

    CastArc = skapi.arc;

    if (CastDebug) {
	CastLog.OutD("Cast_Cast: Wait done, CastComplete [" + CastComplete + "]");
	CastLog.OutD("Cast_Cast: CastState [" + CastState + " / " + castStateText[CastState] + "]");
	CastLog.OutD("Cast_Cast: CastCode  [" + CastCode  + " / " + castCodeText[CastCode] + "]");
	CastLog.OutD("Cast_Cast: CastFail  [" + CastFail  + " / " + castFailText[CastFail] + "]");
	CastLog.OutD("Cast_Cast: CastFail2 [" + CastFail2 + "]");
	CastLog.OutD("Cast_Cast: CastArc   [" + CastArc   + "]");
    } // endif

    if (CastMsgReq && (CastState != castStateDone)) {
	 if (CastDebug) CastLog.OutD("Cast_Cast: Cast message required but not received, spell failure.");
	 CastState = castStateDone;
	 CastCode  = castCodeFail;
	 CastFail  = castFailNoCastMsg;
    } // endif

    // if CastCode == Unknown, then we missed the chat message indicating the spell was cast,
    // the spell cast doesn't produce chat messages (like portal summon spells), or the
    // OnActionComplete event came before the chat message, so lets check the CastArc
    // variable for the spell result
    if (CastCode == castCodeUnknown) {
	if (CastDebug) CastLog.OutD("Cast_Cast: State_Code Unknown, examining CastArc..." );
	if (CastArc == 0) {
	    CastCode = castCodeSuccess;
	} else {
	    CastCode  = castCodeFail;
	    CastFail  = castFailArc;
	    CastFail2 = CastArc;
	} // endif
    } // endif

    // If cast seems to be a success, make sure the spell cast time was normal.  Due to a
    // serverside bug, sometimes the spell doesn't complete casting but the server returns
    // a success code.  Luckily in these cases, the spell cast time is much shorter then a
    // normal spell, so we make sure all spells take a minimum amount of time.
    if (CastCode == castCodeSuccess) {
	if (SlowCast) {
	    var MinTime = this.MinTimeSlow[Level];
	} else {
	    var MinTime = this.MinTimeFast[Level];
	} // endif
	
	if ((CastEndTime-CastStartTime)<MinTime) {
	    // Spell could not possibly be completed in this time, mark spell as failed
	    CastCode = castCodeFail;
	    CastFail = castFailTooShort;
	    if (CastDebug) CastLog.OutD("Cast_Cast: Spell completed too quickly, setting status as failed");
	} // endif
    } // endif

    // If we failed because we were busy, then wait a bit.  Hopefully busy condition will clear
    if ((CastCode == castCodeFail) &&
	((CastFail == castFailBusy) ||
	 ((CastFail == castFailArc) && (CastArc == arcTooBusy)))) {
	skapi.WaitEvent(800, wemFullTimeout);  // Process all events in queue for 800 millseconds
    } // endif

    // Return status code
    if (CastCode == castCodeSuccess) {
	if (CastDebug) CastLog.OutD("Cast_Cast: Success");
	this.FailList[SpellObj.spellid] = 0;
	this.FailCount = 0;
	this.FirstFail = 0;
	return true;
    } else {
	if (CastDebug) CastLog.OutD("Cast_Cast: Failure");
	this.FailList[SpellObj.spellid]++;
	this.FailCount++;
	if ((CastFail == castFailTimeout) || (CastFail == castFailTooShort) || (CastFail == castFailUnknown)) {
	    if (this.FirstFail == 0) this.FirstFail = new Date();

	    // Now see if we have been getting repeated spell failures with the past few seconds
	    if ((new Date() - this.FirstFail) > castFailIntMax) {
	        CastLog.OutD("Cast_Cast: " + this.FailCount + "failures since " + this.FirstFail +
	    		 ". Moving in case we are stuck...");
		this.FailCount = 0;
		this.FirstFail = 0;
		CastFailCnt++;

		// Slide for a bit
		skapi.Keys(cmidSideStepLeft,  kmoDown);
		skapi.Keys(cmidWalkBackwards, kmoDown);
		skapi.WaitEvent(650, wemFullTimeout);
		skapi.Keys(cmidWalkBackwards, kmoUp);
		skapi.Keys(cmidSideStepLeft,  kmoUp);
		skapi.WaitEvent(400, wemFullTimeout);

		// Now hit the escape key.  First save the current selection and reselect
		// when done since the esc key unselected the currently selected creature.
		var CurSelectedAco = skapi.acoSelected;
		skapi.Keys("{ESC}", kmoClick);
		skapi.WaitEvent(300, wemFullTimeout);
		skapi.Keys("{ESC}", kmoClick);
		skapi.WaitEvent(300, wemFullTimeout);
		if (CurSelectedAco != null) skapi.SelectAco(CurSelectedAco);
		CastLog.OutD("Cast_Cast: Unstick attempt ended.");
	    } // endif
	} else {
	    // Spell failed but was cast successfully (ie resist, out of range, fizzle etc)
	    this.FirstFail = 0;
	} // endif

	return false;
    } // endif
} // end func Cast_Cast

function Cast_GetSpellObj(SpellName) {
    if (CastDebug) CastLog.OutD("Cast_GetSpellObj: Begin, SpellName [" + SpellName + "]");

    // Convert the spell name to its spell id
    var SpellId = skapi.SpellidFromSz(SpellName);
    if ((SpellId == null) || (SpellId == 0)) {
    	CastLog.OutD("Cast_GetSpellObj: Unable to translate spell [" + SpellName + "] to a spell id");
    	return null;
    } // endif

    var SpellObj = skapi.SpellInfoFromSpellid(SpellId);
    if (SpellObj == null) {
    	CastLog.OutD("Cast_GetSpellObj: Unable to get spell object for spell id  [" + SpellId + "]");
    	return null;
    } // endif

    // Success!
    return SpellObj;
} // end func Cast_GetSpellObj

function Cast_Know(SpellId) {
    if (CastDebug) CastLog.OutD("Cast_Know: Begin, checking spell id [" + SpellId + "]");

    if ((this.SpellBook[SpellId] == undefined) ||
	(this.SpellBook[SpellId] == null) ||
	(this.SpellBook[SpellId] == false)) {
	if (CastDebug) CastLog.OutD("Cast_Know: We do not know spell!");
	return false;
    } // endif

    if (CastDebug) CastLog.OutD("Cast_Know: We know spell");
    return true;
} // end func Cast_Know

function Cast_ModeCheck(flag) {
    this.ModeCheck = flag;
    return true;
} // end func

function Cast_Abort() {
    CastAbortFlag = true;
    return true;
} // end func

function Cast_FailClear(SpellId) {
    this.FailList[SpellId] = 0;
    return true;
} // end func

function Cast_FailCount(SpellId) {
    return this.FailList[SpellId];
} // end func

function Cast_SetAttTimeout(value) {
    this.AttemptTimeout = value;
    return true;
} // end func

function Cast_SetCompTimout(level,value) {
    this.Timeout[level]=value;
    return true;
} // end func

Cast = new CastObj(false);