// MD_CREATUREOBJ.JS
//
// Author:	Poundboy
// Email:	poundboy@megapathdsl.net
// Version:	1.01
//
//
// Modification History:
//
// Who	  Date	    Version	Description
//
// PB	  09/13/02  1.01	Initial version.  Mostly based on old acscript drainbot
//
//
// Description:
//
// The following routines implement a creature object.   From this object you can determine the
// following about a creature:
//
//	Name
//	Creature Group (Bandlering, Shadow, Golem, etc...)
//	Level
//	Experience points received from killing
//	Maximum Health
//	Maximum Mana
//	Maximum Stamina
//	Magic Defense Skill
//	Weakest element (for war spells)
//
// Requirements:
//
// Requires the following libraries: MD_DEFS.JS, MD_UTIL.JS, MD_LOG.JS
//
// Usage:
//
// Methods:
//
// SetDebug(debugflag,opm)		Turns debugging off/on to opm location
// Init(race_file, cre_file, log_file)	Initializes race and creature structures with info from specified files
// NameToIndex(name)			Returns numeric 'id' of creature with name 'name'
// SetFlags(name,value)			Sets the AttFlag property of creature 'name' with boolean value 'value'
// Dump()				For debugging.  Dumps creature database
//
// 
//
// Creature Properties (loaded from datafile)
//
// Race		Race of creature
// Name		Name of creature
// Level	Creature's level
// MaxHealth	Maximum health
// MaxStamina	Maximum stamina
// MaxMana	Maximum mana
// MagicD	Magic defense skill of creature
// Exp		Experience points gained from killing
// Vuln		Damage type creature is most susceptible to
//
// Properties (set by warbot macro.  JScript doesn't support inheritance so we put these here for use by warbot macro)
//
// Priority	When multiple creatures to select, will select highest priority first
// SelFlag	Can this creature be selected?
// AttFlag	If creature is attacking us, will we fight back?
// YieldFlag	Do we yield this creature?
//
// Usage:
//
// An object of the name 'Creature' is automatically created when you include this file.  Before
// using this object, you must execute the Init method of this object to load the creature data
// from a file.
//
// This object can be used for many purposes such as:
//
// Determining whether or not to attack a creature based on Level, Exp, MagicD,or AttFlag
//
// Determining whether magic yield necessary by looking at MagicD of creature.
//
// Determining best damage type to vuln/war a creature with.
//
// Determining current absolute health of creature by multiplying MaxHealth by % of creature
// health left (from OnTarget event).
//
// Determining whether or not worth draining health, mana, or stamina from a creature by looking
// at MaxHealth, MaxMana, MaxStamina values (should estimate remaining health/mana/stam stats
// by examining results of Drain spells).
//
//
// Modification History
//
// Who	Date	  Version	Description
//
// PBoy	06/23/02  1.01		Initial version
//

// *******************************************
// Global Variables
//
// Constants
//

// Constants for selecting, attacking, and yielding creatures
attAlways	= 1;		// Always perform action
attNever	= 2;		// Never perform action
attAuto		= 0;		// Figure out whether or not to perform action based on creature distance, creature
				// stats, and macroer's magic skill levels.

AttText		= new Array();
AttText[attAlways] = "Always";
AttText[attNever]  = "Never";
AttText[attAuto]   = "Auto";

// Global variables.  Can't put most of these in objects since events fire outside of object context
CreDebug	= false;		// Set to true when Comp logging desired
CreLog		= new LogObj();		// Create log object for comp routines

// ***********************************************************************************************

function RaceRec(RaceName, RaceCode, DefVuln) {
    if (CreDebug) CreLog.OutD("RaceRec: Begin, Name [" + RaceName + "] [" + RaceCode + "] [" + DefVuln + "]");

    // Properties from race file that can NOT be changed by user through decal panels
    this.RaceName	= RaceName;	// Name of race
    this.RaceCode	= RaceCode;	// Species code from assess message

    // Properties from race file that CAN be changed by user through decal panels
    this.DefVuln	= DefVuln;	// Default vuln to use for this race
    
    // Properties initially set by WarBot program and CAN be changed through decal panels
    this.SelFlag	= attAuto;	// By default this race is selectable if in range and...
    this.AttFlag	= attAuto;	// we will attack this race if they enter our danger radius
    
    // Race properties managed by warbot program only
    this.NList		= new Array();	// Creature list for this race in database order.  Used to build decal schemas.
    this.NListCnt	= 0;		// Array entries are indexes into Creature.Info array.
    this.AList		= new Array();	// Creature list for this race in alpha order.  Used to build decal schemas.
    this.AListCnt	= 0;		// Array entries are indexes into Creature.Info array.
    
    // Save user changable properties.  Used to reset those properties to original values and to determine
    // what has changed when writing user changed properties to creature config file.
    this.OrigDefVuln	= this.DefVuln;
    this.OrigSelFlag	= this.SelFlag;
    this.OrigAttFlag	= this.AttFlag;

    if (CreDebug) CreLog.OutD("RaceRec: End");
} // end func

function CreatureRec(RaceCode, Name, Vuln1, Vuln2, Vuln3, Health, Mana, Stamina, MagicD, Level, Exp,
		     NoDrainH, NoDrainS, NoDrainM ) {
    if (CreDebug) CreLog.OutD("CreatureRec: Begin, Name [" + Name + "]");

    // Following properties read from creature file and can NOT be changed via the decal panels
    this.RaceCode	= RaceCode;	// Numeric code of race
    this.Name		= Name;		// Creatures Name
    this.Health		= Health;	// Maximum health
    this.Mana		= Mana;		// Maximum mana
    this.Stamina	= Stamina;	// Maximum stamina
    this.MagicD		= MagicD;	// Magic Defense
    this.Level		= Level;	// Level
    this.Exp		= Exp;		// Experience points gained from killing.
 
    // Following properties read from creature file and CAN be changed via the decal panels   
    this.Vuln1		= Vuln1;	// Element creature is least resistant to
    this.Vuln2		= Vuln2;	// Secondary vuln element
    this.Vuln3		= Vuln3;	// Tertiary vuln element
    this.NoDrainHealth	= NoDrainH;	// If true, then don't drain health from this creature
    this.NoDrainStamina	= NoDrainS;	// If true, then don't drain stamina from this creature
    this.NoDrainMana	= NoDrainM;	// If true, then don't drain mana from this creature

    // Following properties initially set by WarBot program but can be changed via the decal panels
    this.Priority	= Exp;		// Default priority is the creature's experience points.
    this.SelFlag	= attAuto;	// By default this creature is selectable if in range and ...
    this.AttFlag	= attAuto;	// we will attack it if it enters our danger radius and ...
    this.YieldFlag	= attAuto;	// will yield it if necessary and ...
    this.VulnFlag	= attAuto;	// will vuln it if necessary.
    this.WarSpell	= attSpellAuto;	// Auto = let macro decided what type of war spell
    this.NoStepDown	= false;	// If set to true, never step down life or war spells for this creature.
    this.DangerFlag	= false;	// If set to true, then we will try to kill this creature as fast as
    					// possible and before non-dangerous creatures.

    // Properties private to warbot
    this.Unknown	= false;	// Set to true when creature was not in creature database (ie was auto-added)
    this.Assessed	= false;	// Set to true when a creature has been successfully assessed.
    					// Used to gather stats from a creature.
    this.AssessCnt	= 0;		// # of times assess has been tried
    this.OrigRace	= RaceCode;	// Original species code if race unknown (RaceCode will be set to 0).		
    this.BadRace	= false;	// Set to true when an exception logged to the creature exception
    this.BadStats	= false;	// file for this creature .  Used when there is a creature-race
    					// mismatch or a creatures stats are wrong. Since we don't want to
    					// flood the log with the same entries, these vars are used to log
    					// only the first occurance.

    // Save original settings in case user wants to reset all user settable properties.  Also used when
    // saving creature configuration to determine what is different then the default.
    this.OrigVuln1		= this.Vuln1;
    this.OrigVuln2		= this.Vuln2;
    this.OrigVuln3		= this.Vuln3;
    this.OrigPriority		= this.Priority;
    this.OrigSelFlag		= this.SelFlag;
    this.OrigAttFlag		= this.AttFlag;
    this.OrigYieldFlag		= this.YieldFlag;
    this.OrigVulnFlag		= this.VulnFlag;
    this.OrigWarSpell		= this.WarSpell;
    this.OrigNoDrainHealth	= this.NoDrainHealth;
    this.OrigNoDrainStamina	= this.NoDrainStamina;
    this.OrigNoDrainMana	= this.NoDrainMana;
    this.OrigNoStepDown		= this.NoStepDown;
    this.OrigDangerFlag		= this.DangerFlag;
    
    if (CreDebug) CreLog.OutD("CreatureRec: End");
} // end func

function CreatureObj(debugflg) {
    if ((debugflg==undefined) || (debugflg==null)) debugflg = false;
    CreDebug = debugflg // Outputs debugging info to debug destination if set

    // Variables
    this.InitFlag	= false;		// Set to true when creature list read from file
    this.LogEnabled	= false;		// Set to true when logging
    this.LogFile	= null;			// Log file (if specified) for unknown creatures/races and db errors
    this.CreCnt		= 0;			// # of creatures in Info array
    this.Info		= new Array();		// Array of CreatureRecs
    this.RaceMax	= 0;			// Maximum race entry code used
    this.RaceMin	= 9999;			// Minimum race code
    this.RaceInfo	= new Array();		// Array of RaceRecs indexed by species code
    this.RaceDecal	= new Array();		// Decal display order of race records.  Same as order in race.csv file.
    this.RaceDecalCnt	= 0;			// Number of entries in RaceAlpha list

    // Name to index translations
    this.Name		= new ActiveXObject("Scripting.Dictionary");
    this.Race		= new ActiveXObject("Scripting.Dictionary");
    
    // Unknown race tracking
    this.UnknownRace	= new Array();		// Unknown races indexed by species code offset

    //Public methods
    this.SetDebug	= Creature_SetDebug;
    this.Init		= Creature_Init;
    this.NameToIndex	= Creature_NameToIndex;
    this.RaceToIndex	= Creature_RaceToIndex;
    this.FindSpot	= Creature_FindSpot;
    this.AddDecal	= Creature_AddDecal;
    this.SetCre		= Creature_SetCre;
    this.SetCre2	= Creature_SetCre2;
    this.ResetCre	= Creature_ResetCre;
    this.AddUnknownCre	= Creature_AddUnknownCre;
    this.DumpCre	= Creature_DumpCre;
    this.SetRace	= Creature_SetRace;
    this.SetRace2	= Creature_SetRace2;
    this.ResetRace	= Creature_ResetRace;
} // end func

function Creature_SetDebug(debugflag) {
    CreDebug = debugflag;
    return true;
} // end func


function Creature_Init(race_file, cre_file, log_file) {
    if (CreDebug) CreLog.OutD("Creature_Init: Begin, racefile [" + race_file + "] crefile [" + cre_file +
    			      "] logfile [" + log_file + "]");

    // See if a log file is defined.  Used to write unknown races/creatures and database errors
    if (log_file != undefined) {
	this.LogEnabled = true;
	this.LogFile = new LogFileObj(log_file, ForAppending, true);
	this.LogFile.WriteLineD(" ");
	this.LogFile.WriteLineD("Log file opened.");
    } // endif

    // Create a file system object
    this.myFileSysObj	= new ActiveXObject("Scripting.FileSystemObject");

    // Read in race file
    CreLog.OutD("Creature_Init: Processing race file [" + race_file + "]...");
    this.TextFileObj	= this.myFileSysObj.OpenTextFile(race_file, 1, false);

    // Race entry for array position 0.  Used to set defaults for unknown species
    this.RaceInfo[0] = new RaceRec("Unknown", 0, damNum['Blade']);
    this.RaceDecal[this.RaceDecalCnt] = 0;
    this.RaceDecalCnt++;

    while (!(this.TextFileObj.AtEndOfStream)) {
	// now read lines until EOF
	line = this.TextFileObj.ReadLine();
        if (CreDebug) CreLog.OutD("Creature_Init (race): Line [" + line + "]");

	split = line.split(",");
	// for (var x=0; x<split.length; x++) CreLog.OutD( x + " [" + split[x] + "]");

	var RaceName	= Util.Trim(split[0]);
	var RaceCode	= parseInt(split[1]);
	var Vuln1	= Util.Trim(split[2]);

	if (CreDebug) {
	    CreLog.OutD("RaceName [" + RaceName + "]");
	    CreLog.OutD("RaceCode [" + RaceCode	+ "]");
	    CreLog.OutD("Vuln1    [" + Vuln1 	+ "]");
	} // endif
	
	if ((this.RaceInfo[RaceCode] != undefined) && (this.RaceInfo[RaceCode] != null)) {
	    CreLog.OutD("Creature_Init: RaceCode [" + RaceCode + "] already exists skipping race [" + RaceName + "] ...");
	    continue;
	} //endif

	// Convert primary vuln text into equivalent numeric code.  If vuln unrecognized then log error
	// message and skip race
	Vuln1Num = damNum[Vuln1];
	if ((Vuln1Num == undefined) || (Vuln1Num == null) || (Vuln1Num == 0)) {
	    CreLog.OutD("Creature_Init: Primary vuln name [" + Vuln1 + "] unrecognized for race [" + RaceName + "] skipping...");
	    continue;
	} // endif

	// Add race name to Race scripting object (used for race name to index translations)
	if (this.Race.Exists(RaceName)) {
	    CreLog.OutD("Error.  Race name already exists [" + RaceCode + " /" + RaceName + "] exists, skipping...");
	    continue;
	} // endif
	this.Race.Add(RaceName,RaceCode);

	// Add this race to the database
	this.RaceInfo[RaceCode] = new RaceRec(RaceName, RaceCode, Vuln1Num);
	
	// Add this record to array that holds decal display order of race
	this.RaceDecal[this.RaceDecalCnt] = RaceCode;
    	this.RaceDecalCnt++;
	
	// Keep track of the minimum and maximum race codes in case we need to go through the whole list
	if (RaceCode < this.RaceMin) this.RaceMin = RaceCode;
	if (RaceCode > this.RaceMax) this.RaceMax = RaceCode;
    } // end while
    
    if (this.RaceMin == 9999) this.RaceMin = 0;	

    // Close race file
    this.TextFileObj.Close();
    CreLog.OutD("Creature_Init: Race File Processed");

    // Read in creature list file and populate create record array. First open file
    CreLog.OutD("Creature_Init: Processing file [" + cre_file + "]...");
    this.TextFileObj = this.myFileSysObj.OpenTextFile(cre_file, 1, false);

    // We don't use creature index 0.  Reserved for unknown creatures.
    this.CreCnt++;

    while (!(this.TextFileObj.AtEndOfStream)) {
	// now read lines until EOF
	line = this.TextFileObj.ReadLine();
        if (CreDebug) CreLog.OutD("Creature_Init (cre): Line [" + line + "]");

	split = line.split(",");
	// for (var x=0; x<split.length; x++) CreLog.OutD( x + " [" + split[x] + "]");

	var RaceCode	= parseInt(split[0]);
	var Name	= Util.Trim(split[1]);
	var Level	= parseInt(split[2]);
	var Health	= parseInt(split[3]);
	var Stamina	= parseInt(split[4]);
	var Mana	= parseInt(split[5]);
	var MagicD	= parseInt(split[6]);
	var Exp		= parseInt(split[7]);
	var Vuln1	= Util.Trim(split[8]);
	var Vuln2	= Util.Trim(split[9]);
	var Vuln3	= Util.Trim(split[10]);
	var NoDrainH	= Util.Trim(split[11]);
	var NoDrainS	= Util.Trim(split[12]);
	var NoDrainM	= Util.Trim(split[13]);
	// var Comment	= Util.Trim(split[14]);

	if (CreDebug) {
	    CreLog.OutD("RaceCode [" + RaceCode	+ "]");
	    CreLog.OutD("Name     [" + Name 	+ "]");
	    CreLog.OutD("Level    [" + Level	+ "]");
	    CreLog.OutD("Health   [" + Health 	+ "]");
	    CreLog.OutD("Stamina  [" + Stamina  + "]");
	    CreLog.OutD("Mana     [" + Mana  	+ "]");
	    CreLog.OutD("MagicD   [" + MagicD   + "]");
	    CreLog.OutD("Exp      [" + Exp	+ "]");
	    CreLog.OutD("Vuln1    [" + Vuln1 	+ "]");
	    CreLog.OutD("Vuln2    [" + Vuln2	+ "]");
	    CreLog.OutD("Vuln3    [" + Vuln3	+ "]");
	    CreLog.OutD("NoDrainH [" + NoDrainH	+ "]");
	    CreLog.OutD("NoDrainS [" + NoDrainS	+ "]");
	    CreLog.OutD("NoDrainM [" + NoDrainM	+ "]");
	} // endif
	
	if ((this.RaceInfo[RaceCode] == undefined) || (this.RaceInfo[RaceCode] == null)) {
	    CreLog.OutD("Creature_Init: RaceCode [" + RaceCode + "] unrecognized for creature [" + Name + "] skipping...");
	    continue;
	} //endif

	// Convert primary vuln text into equivalent numeric code.  If vuln unrecognized then log error
	// message to log and skip creature.
	Vuln1Num = damNum[Vuln1];
	if ((Vuln1Num == undefined) || (Vuln1Num == null) || (Vuln1Num == 0)) {
	    CreLog.OutD("Creature_Init: Primary vuln name [" + Vuln1 + "] unrecognized for creature [" + Name + "] skipping...");
	    continue;
	} // endif

	// Convert secondary vuln text into equivalent numeric code.   Not required.
	if ((Vuln2 == undefined) || (Vuln2 == null) || (Vuln2 == "")) {
	    Vuln2Num = 0;
	} else {
	    Vuln2Num = damNum[Vuln2];
	    if ((Vuln2Num == undefined) || (Vuln2Num == null) || (Vuln2Num == 0)) {
		CreLog.OutD("Creature_Init: Secondary vuln name [" + Vuln2 + "] unrecognized for creature [" + Name + "]");
		Vuln2Num = 0;
	    } // endif
	} // endif

	// Convert tertiary vuln text into equivalent numeric code.   Not required.
	if ((Vuln3 == undefined) || (Vuln3 == null) || (Vuln3 == "")) {
	    Vuln3Num = 0;
	} else {
	    Vuln3Num = damNum[Vuln3];
	    if ((Vuln3Num == undefined) || (Vuln3Num == null) || (Vuln3Num == 0)) {
		CreLog.OutD("Creature_Init: Tertiary vuln name [" + Vuln3 + "] unrecognized for creature [" + Name + "]");
		Vuln3Num = 0;
	    } // endif
	} // endif
	
	if (NoDrainH == "false") {
	    NoDrainH = false;
	} else {
	    if (NoDrainH == "true") {
	    	NoDrainH = true;
	    } else {
	    	// Bad entry
	    	 CreLog.OutD("Creature_Init: NoDrainHealth flag [" + NoDrainH + "] unrecognized for creature [" + Name +
	    	 	     "] skipping...");
	   	continue;
	    } // endif
	} //endif

	if (NoDrainS == "false") {
	    NoDrainS = false;
	} else {
	    if (NoDrainS == "true") {
	    	NoDrainS = true;
	    } else {
	    	// Bad entry
	    	 CreLog.OutD("Creature_Init: NoDrainStamina flag [" + NoDrainS + "] unrecognized for creature [" + Name +
	    	 	     "] skipping...");
	   	continue;
	    } // endif
	} //endif

	if (NoDrainM == "false") {
	    NoDrainM = false;
	} else {
	    if (NoDrainM == "true") {
	    	NoDrainM = true;
	    } else {
	    	// Bad entry
	    	 CreLog.OutD("Creature_Init: NoDrainMana flag [" + NoDrainM + "] unrecognized for creature [" + Name +
	    	 	     "] skipping...");
	   	continue;
	    } // endif
	} //endif

	// Add creature name to Name object (used for name to index translations)
	if (this.Name.Exists(Name)) {
	    CreLog.OutD("Creature_Init: Creature name already exists [" + Name + "] exists, skipping...");
	    continue;
	} else {
	    this.Name.Add(Name,this.CreCnt);
	} // endif

	// Add to CreatureRec list
	this.Info[this.CreCnt] = new CreatureRec( RaceCode, Name, Vuln1Num, Vuln2Num, Vuln3Num, Health, Mana,
					          Stamina, MagicD, Level, Exp, NoDrainH, NoDrainS, NoDrainM );

	// Now add this creature to the creature list for this race
	this.AddDecal(RaceCode, this.CreCnt);

	// Increment count of creatures
	this.CreCnt++;	

    } // end while

    // Close file
    this.TextFileObj.Close();
    CreLog.OutD("Creature_Init: Creature File Processed");

    // All done, set init flag to true and return
    this.InitFlag = true;

    return true;
} // end func
    
function Creature_AddDecal(RaceCode, CreIndex) {
    with (this.RaceInfo[RaceCode]) {
	// First do the database order list
	NList[NListCnt] = CreIndex;
	NListCnt++;
	    
	// Now do the alphabetic list.  First find the offset to put the entry.  Since the
	// list is in alpha order, we can do a binary search for the offset.
	var Offset = this.FindSpot(RaceCode, CreIndex);
	// Chat.Screen("OFFSET [" + Offset + "]");
	    
	// If the offset is in the middle of the list, then shift all the entries down by 1
	if (Offset < AListCnt) {
	    for (var x=AListCnt; x>Offset; x--) AList[x] = AList[x-1];
	} // endif
	AList[Offset] = CreIndex;
	AListCnt++;    
    } // end with
} // end function


function Creature_FindSpot(RaceCode, CreIndex) {
    if (this.RaceInfo[RaceCode].AListCnt == 0) return 0;
    
    with (this.RaceInfo[RaceCode]) {
	found = false;
	var begin = 0;
	var end = AListCnt-1;
	while (!found) {
	    if (this.Info[CreIndex].Name < this.Info[AList[begin]].Name) return begin;
	    if (this.Info[CreIndex].Name > this.Info[AList[end]].Name) return end+1;
	    if (begin==end) {
	    	// Spot is either before or after the convergence element
	    	if (this.Info[CreIndex].Name < this.Info[AList[begin]].Name)
		    return begin;
		else
		    return begin+1;
	    } // endif
	    var test = Math.round((begin+end)/2);
	    if (this.Info[CreIndex].Name < this.Info[AList[test]].Name)
		end = test - 1;
	    else
	    	begin = test + 1;
	} // end while
    } // end with
} // end func

// Translate name to index
function Creature_NameToIndex(name) {
    if (name == undefined) {
    	if (CreDebug) CreLog.OutD("Creature_NameToIndex: Name undefined");
    	return 0;
    } // endif

    if (name == null) {
    	if (CreDebug) CreLog.OutD("Creature_NameToIndex: Name null");
    	return 0;
    } // endif

    if (name == "") {
    	if (CreDebug) CreLog.OutD("Creature_NameToIndex: Name blank");
    	return 0;
    } // endif

    if (CreDebug) CreLog.OutD("Creature_NameToIndex: Name lookup [" + name + "]" );

    // If name exists, then return it, else return 0
    if (this.Name.Exists(name)) {
    	if (CreDebug) CreLog.OutD("Creature_NameToIndex: Name exists, getting associated item...");
    	var index = this.Name.Item(name);
    	if (CreDebug) CreLog.OutD("Creature_NameToIndex: Item lookup [" + index + "]");
    	return parseInt(index);
    } // endif

    // Name doesn't exist
    return 0;	// Return index of default creature
} // end func


// Translate name to index
function Creature_RaceToIndex(race) {
    if (race == undefined) {
    	if (CreDebug) CreLog.OutD("Creature_RaceToIndex: Race undefined");
    	return 0;
    } // endif

    if (race == null) {
    	if (CreDebug) CreLog.OutD("Creature_RaceToIndex: Race null");
    	return 0;
    } // endif

    if (race == "") {
    	if (CreDebug) CreLog.OutD("Creature_RaceToIndex: Race blank");
    	return 0;
    } // endif

    if (CreDebug) CreLog.OutD("Creature_RaceToIndex: Race lookup [" + race + "]" );

    // If race exists, then return it, else return 0
    if (this.Race.Exists(race)) {
    	if (CreDebug) CreLog.OutD("Creature_RaceToIndex: Race exists, getting associated item...");
    	var index = this.Race.Item(race);
    	if (CreDebug) CreLog.OutD("Creature_RaceToIndex: Race index [" + index + "]");
    	return parseInt(index);
    } // endif

    // Race doesn't exist
    return 0;	// Return index of default creature
} // end func

// Old function
function Creature_SetCre(Name, SelFlag) {
    var Index = this.NameToIndex(Name);
    if (Index==0) {
    	CreLog.OutD("Creature_SetCre: Unknown creature [" + Name + "]");
    	return false;
    } // endif

    // If flag specified, then set it
    this.Info[Index].SelFlag = SelFlag;
    return true;
} // end function


// Current function that will be referenced when config file is next written
function Creature_SetCre2(Name, Vuln1, Vuln2, Vuln3, Priority, SelFlag, AttFlag, YieldFlag, VulnFlag,  
			 WarSpell, NoDrainHealth, NoDrainStamina, NoDrainMana, NoStepDown, DangerFlag) {
    var Index = this.NameToIndex(Name);
    if (Index == 0) {
    	CreLog.OutD("Creature_SetCre: Unknown creature [" + Name + "]");
    	return false;
    } // endif

    this.Info[Index].Vuln1 		= Vuln1;
    this.Info[Index].Vuln2		= Vuln2;
    this.Info[Index].Vuln3		= Vuln3;
    this.Info[Index].Priority		= Priority;
    this.Info[Index].SelFlag		= SelFlag;
    this.Info[Index].AttFlag		= AttFlag;
    this.Info[Index].YieldFlag		= YieldFlag;
    this.Info[Index].VulnFlag		= VulnFlag;
    this.Info[Index].WarSpell		= WarSpell;
    this.Info[Index].NoDrainHealth	= NoDrainHealth;
    this.Info[Index].NoDrainStamina	= NoDrainStamina;
    this.Info[Index].NoDrainMana	= NoDrainMana;
    this.Info[Index].NoStepDown		= NoStepDown;
    this.Info[Index].DangerFlag		= DangerFlag;
    return true;
} // end function


function Creature_ResetCre(CreCode) {
    // If no parameter passed to this function, then reset all creatures, else
    // just reset the specified creature.
    if ((CreCode == undefined) || (CreCode == null)) {
    	var begin = 1;			// Array position 0 not used
    	var end   = this.CreCnt-1;	// Creature count is 1 higher than max array position used
    } else {
    	var begin = CreCode;
    	var end   = CreCode;
    } // endif
    
    for (var x=begin; x<=end; x++) {
	this.Info[x].Vuln1		= this.Info[x].OrigVuln1;
	this.Info[x].Vuln2		= this.Info[x].OrigVuln2;
	this.Info[x].Vuln3		= this.Info[x].OrigVuln3;
	this.Info[x].Priority		= this.Info[x].OrigPriority;
	this.Info[x].SelFlag		= this.Info[x].OrigSelFlag;
	this.Info[x].AttFlag		= this.Info[x].OrigAttFlag;
	this.Info[x].YieldFlag		= this.Info[x].OrigYieldFlag;
	this.Info[x].VulnFlag		= this.Info[x].OrigVulnFlag;
	this.Info[x].WarSpell		= this.Info[x].OrigWarSpell;
	this.Info[x].NoDrainHealth	= this.Info[x].OrigNoDrainHealth;
	this.Info[x].NoDrainStamina	= this.Info[x].OrigNoDrainStamina;
	this.Info[x].NoDrainMana	= this.Info[x].OrigNoDrainMana;
	this.Info[x].NoStepDown		= this.Info[x].OrigNoStepDown;
	this.Info[x].DangerFlag		= this.Info[x].OrigDangerFlag;
    } // endif
} // end func

function Creature_AddUnknownCre(CreAco, Verbose) {
    if (CreDebug) CreLog.OutD("Creature_AddUnknownCre: Begin CreName [" + CreAco.szName + "] Verbose [" + Verbose + "]");
    
    // First get some more information on the creature by assessing it.  Even if we can't assess it we always
    // get basic creature information which includes species (race) code, level, max health, and current health.
    // Assess sets following global variables: CreCbi, CreCai, CreChi
    var Status = Assess.AssessCre(CreAco);
    if (!Status) {
	// Access attempt failed ie didn't get success or failure data back
	if (CreDebug) CreLog.OutD("Creature_AddUnknownCre: Failed to assess");
	return 0;
    } // endif

    var MonIndex = this.CreCnt;
    this.CreCnt++;
    this.Name.Add(CreAco.szName, MonIndex);
	    
    var Level     = 0;
    var HealthMax = 0;
    var RaceCode  = 0;
    var OrigRace  = 0;
    
    if (CreCbi) {
	Level     = CreCbi.lvl;
	HealthMax = CreCbi.healthMax;
	RaceCode  = CreCbi.species;
	OrigRace  = RaceCode;
	
	// See if species known
	if ((this.RaceInfo[RaceCode] == undefined) || (this.RaceInfo[RaceCode] == null)) {
	    // Unknown species
	    RaceCode = 0;
	    
	    // See if this is the first time we have seen this unknown race
	    if (this.UnknownRace[OrigRace] == undefined) {
		// First time we have seen this race
		this.UnknownRace[OrigRace] = true;
		if (this.LogEnabled) {
		    this.LogFile.WriteLineD(" ");
		    this.LogFile.WriteLineD("Unknown race detected, species code [" + OrigRace + "]");
		} // endif
	    } // endif
	} // endif
    } // endif

    var StaminaMax = 0;
    var ManaMax = 0;
    if (CreCai) {
	StaminaMax = CreCai.staminaMax;
	ManaMax = CreCai.manaMax;
    } // endif
	    
    Vuln = this.RaceInfo[RaceCode].DefVuln;

    // Add to CreatureRec list
    this.Info[MonIndex] = new CreatureRec(RaceCode, CreAco.szName, Vuln, 0, 0, HealthMax, ManaMax,
					  StaminaMax, 0, Level, 0, false, false, false);
    this.Info[MonIndex].Unknown   = true;
    this.Info[MonIndex].AssessCnt = 1;
    this.Info[MonIndex].OrigRace  = OrigRace;

    if (CreCbi && CreCai) {
	this.Info[MonIndex].Assessed = true;
	if (CreDebug) CreLog.OutD("Creature_AddUnknownCre: Assess succeeded.");
    } else {
	this.Info[MonIndex].Assessed = false;
	if (CreDebug) CreLog.OutD("Creature_AddUnknownCre: Assess failed.");
    } // endif

    // If logging enabled, then log this creatures stats to the creature file
    if (this.LogEnabled) {
	this.LogFile.WriteLineD(" ");
	this.LogFile.WriteLineD("Unknown creature detected.");
	this.LogFile.WriteLineD("Name    [" + CreAco.szName + "]");
	this.LogFile.WriteLineD("Species [" + OrigRace      + "]");
	this.LogFile.WriteLineD("Level   [" + Level         + "]");
	this.LogFile.WriteLineD("Health  [" + HealthMax     + "]");
	this.LogFile.WriteLineD("Mana    [" + ManaMax       + "]");
	this.LogFile.WriteLineD("Stamina [" + StaminaMax    + "]");
    } // endif
 
    // Now add creature to decal lists
    this.AddDecal(RaceCode, MonIndex);
    
    return MonIndex;

} // end func


// Old function
function Creature_SetRace(Name, SelFlag) {
   var Index = this.RaceToIndex(Name);
    if (Index==0) {
    	CreLog.OutD("Creature_SetRace: Unknown race [" + Name + "]");
    	return false;
    } // endif
    this.RaceInfo[Index].SelFlag = SelFlag;
    return true;
} // end func

// Current function that will be referenced when config file is next written
function Creature_SetRace2(Name, DefVuln, SelFlag, AttFlag) {
   var Index = this.RaceToIndex(Name);
    if (Index == 0) {
    	CreLog.OutD("Creature_SetRace2: Unknown race [" + Name + "]");
    	return false;
    } // endif
    this.RaceInfo[Index].DefVuln = DefVuln;
    this.RaceInfo[Index].SelFlag = SelFlag;
    this.RaceInfo[Index].AttFlag = AttFlag;
    return true;
} // end func

function Creature_ResetRace(RaceCode) {
    // If no parameter passed to this function, then reset all races, else
    // just reset the specified race.
    if ((RaceCode == undefined) || (RaceCode == null)) {
    	var begin = 0;
    	var end   = this.RaceMax;
    } else {
    	var begin = RaceCode;
    	var end   = RaceCode;
    } // endif

    // Chat.Screen("ResetRace: begin [" + begin + "] end [" + end + "]");
    for (var x=begin; x<=end; x++) {
    	if ((this.RaceInfo[x] != undefined) && (this.RaceInfo[x] != null)) {
	    this.RaceInfo[x].DefVuln = this.RaceInfo[x].OrigDefVuln;
	    this.RaceInfo[x].SelFlag = this.RaceInfo[x].OrigSelFlag;
	    this.RaceInfo[x].AttFlag = this.RaceInfo[x].OrigAttFlag;
	} // endif
    } // end for
    
    return true;
} // end func

// Need to finish this
function Creature_DumpCre() {
    for (x=1; x<this.CreCnt; x++) {
    	with (this.Info[x]) {
	    CreLog.OutD( Name + " " + Group );
	    CreLog.OutD( Vuln + " " + Health + " " + Mana + " " + Stamina + " " + MagicD );
	    CreLog.OutD( Level + " " + Exp + " " + Priority + " " + SelFlag + " " + AttFlag + " " + YieldFlag );
	} // end with
    } // endif
    return;
} // endif

Creature = new CreatureObj(false);


 