// MD_DAMAGEOBJ.JS
//
// Author:	Poundboy
// Email:	poundboy@megapathdsl.net
// Version:	1.01
//
// Description:
//
// The following object/records implement structures that make it easy to find the spell id that
// affects a particular damage type for war, life and item spells.   When this file is included
// in a project, the object 'Damage' is created automatically.
//
// For example, when used in conjunction with the Creature and Cast objects, you could do something
// like this (assume TargetAco is the aco of the creature we want to kill);
//
// maxLifeSpell = 7;				// Set highest life spell we can cast
// maxWarSpell  = 6;				// Set highest war spell we can cast
// cre_id = Creature.NameToId(TargetAco.szName)	// Get creatures id from its aco name
// Weakness = Creature[cre_id].Vuln1		// Get element creature least resistant to
// Cast.ById ( Damage[Weakness].Vuln[maxLifeSpell], TargetAco );	// Cast vuln
// Cast.ById ( Damage[Weakness].Bolt[maxWarSpell], TargetAco );		// Cast war spell
//
//
// Prerequisites:
//
// Requires the following libraries: MD_DEFS.JS, MD_UTIL.JS, MD_LOGOBJ.JS
//
// Methods:
//
// Dump()		Used for debugging.  Dumps Damage records/structures to log device
//
//
// Properties:
//
//  The following properties are supported:
//
//
// War spells:
//
// Bolt		array of war spells indexed by spell level
// Streak	array of war spells indexed by spell level
// Wall		Spell id of the wall spell for this damage type
// Ring		Spell id of the ring spell for this damage type
// Rain		Spell id of the ring spell for this damage type
//  
// Protects/Banes:
//
// ProtSelf	array of life self protect spells indexed by spell level
// ProtOther	array of life other protect spells indexed by spell level
// Bane		array of item bane spells indexed by spell level
//   
// Debuffs:
//
// Yield	
// Vuln		array of life vuln spells indexed by spell level
// Lure		array of item debuff spells indexed by spell level
//
//
// Modification History
//
// Who	Date	  Version	Description
//
// PB	06/23/02  1.01		Initial version
//
// *******************************************
//
// Global variables/constants
DamLog		= new LogObj();		// Log object used to output to console/chat window
DamDebug	= false;		// If set to true, will output debugging information

// Group all the spells that affect a damage type. Record will be filled in later
function DamageRec() {

    // Offensive spells
    this.Bolt   	= 0;
    this.Streak 	= 0;
    this.Volley 	= 0;
    this.Blast	 	= 0;
    this.Arc		= 0;
    this.Wall		= 0;
    this.Ring		= 0;
    this.Rain		= 0;
    
    // Protects against this damage type
    this.ProtSelf	= 0;
    this.ProtOther	= 0;
    this.Bane		= 0;
    
    // Debuffs for this damage type
    this.Vuln		= 0;
    this.Lure		= 0;
} // end DamageRec

function DamageObj(debugflg) {
    if ((debugflg==undefined) || (debugflg==null)) debugflg = false;
    DamDebug = debugflg;	// Outputs debugging info to debug destination if set
	
    // Define array for spell info
    this.Info	= new Array();
    
    // Create the damage rec for each damage type
    for (var x=1; x<9; x++) this.Info[x] = new DamageRec();
    
    // Now fill in the damage records for each damage type.  Note the level of the spell maps
    // directly into the arrays, so the 0 at the begining of the array lists is for the level 0
    // spell which obviously doesn't exist.
    
    this.Info[damBlade].Bolt		= new Array(0,92,93,94,95,96,97,2146);
    this.Info[damBlade].Streak		= new Array(0,1826,1827,1828,1829,1830,1831,2147);
    this.Info[damBlade].Volley		= new Array(0,0,0,151,152,153,154,0);
    this.Info[damBlade].Blast		= new Array(0,0,0,123,124,125,126,0);
    this.Info[damBlade].Arc		= new Array(0,2753,2754,2755,2756,2757,2758,2759);  
    this.Info[damBlade].Wall		= 1840;
    this.Info[damBlade].Ring		= 1784;
    this.Info[damBlade].Rain		= 1833; 
    this.Info[damBlade].ProtSelf 	= new Array(0,1109,1110,1111,1112,1113,1114,2151);
    this.Info[damBlade].ProtOther	= new Array(0,1115,1116,1117,1118,1119,1120,2150);
    this.Info[damBlade].Vuln		= new Array(0,1127,1128,1129,1130,1131,1132,2164);
    this.Info[damBlade].Bane		= new Array(0,37,1558,1559,1560,1561,1562,2094);
    this.Info[damBlade].Lure		= new Array(0,38,1553,1554,1555,1556,1557,2095);
    
    this.Info[damPierce].Bolt		= new Array(0,86,87,88,89,90,91,2132);
    this.Info[damPierce].Streak		= new Array(0,1802,1803,1804,1805,1806,1807,2133);
    this.Info[damPierce].Volley		= new Array(0,0,0,147,148,149,150);
    this.Info[damPierce].Blast		= new Array(0,0,0,119,120,121,122);
    this.Info[damPierce].Arc		= new Array(0,2718,2719,2720,2721,2722,2723,2724);   
    this.Info[damPierce].Wall		= 1842;
    this.Info[damPierce].Ring		= 1786;
    this.Info[damPierce].Rain		= 1835;
    this.Info[damPierce].ProtSelf 	= new Array(0,1133,1134,1135,1136,1137,1138,2161);
    this.Info[damPierce].ProtOther	= new Array(0,1139,1140,1141,1142,1143,1144,2160);
    this.Info[damPierce].Vuln		= new Array(0,1151,1152,1153,1154,1155,1156,2174);
    this.Info[damPierce].Bane		= new Array(0,1569,1570,1571,1572,1573,1574,2113);
    this.Info[damPierce].Lure		= new Array(0,1563,1564,1565,1566,1567,1568,2114);
    
    this.Info[damBludgeon].Bolt		= new Array(0,64,65,66,67,68,69,2144);
    this.Info[damBludgeon].Streak	= new Array(0,1820,1821,1822,1823,1824,1825,2145);
    this.Info[damBludgeon].Volley	= new Array(0,0,0,131,132,133,134,0);
    this.Info[damBludgeon].Blast	= new Array(0,0,0,103,104,105,106,0);
    this.Info[damBludgeon].Arc		= new Array(0,2746,2747,2748,2749,2750,2751,2752);    
    this.Info[damBludgeon].Wall		= 1845;
    this.Info[damBludgeon].Ring		= 1789;
    this.Info[damBludgeon].Rain		= 1836;
    this.Info[damBludgeon].ProtSelf 	= new Array(0,1018,1019,1020,1021,1022,1023,2153);
    this.Info[damBludgeon].ProtOther	= new Array(0,1024,1025,1026,1027,1028,1029,2152);
    this.Info[damBludgeon].Vuln		= new Array(0,1048,1049,1050,1051,1052,1053,2166);
    this.Info[damBludgeon].Bane		= new Array(0,1511,1512,1513,1514,1515,1516,2098);
    this.Info[damBludgeon].Lure		= new Array(0,1505,1506,1507,1508,1509,1510,2099);
    
    this.Info[damCold].Bolt		= new Array(0,28,70,71,72,73,74,2136);
    this.Info[damCold].Streak		= new Array(0,1808,1809,1810,1811,1812,1813,2137);
    this.Info[damCold].Volley		= new Array(0,0,0,135,136,137,138,0);
    this.Info[damCold].Blast		= new Array(0,0,0,107,108,109,110,0);
    this.Info[damCold].Arc		= new Array(0,2725,2726,2727,2728,2729,2730,2731);    
    this.Info[damCold].Wall		= 1843;
    this.Info[damCold].Ring		= 1787;
    this.Info[damCold].Rain		= 1836; 
    this.Info[damCold].ProtSelf 	= new Array(0,1030,1031,1032,1033,1034,1035,2155);
    this.Info[damCold].ProtOther	= new Array(0,1036,1037,1038,1039,1040,1041,2154);
    this.Info[damCold].Vuln		= new Array(0,1060,1061,1062,1063,1064,1065,2168);
    this.Info[damCold].Bane		= new Array(0,1523,1524,1525,1526,1527,1528,2104);
    this.Info[damCold].Lure		= new Array(0,1517,1518,1519,1520,1521,1522,2105);

    this.Info[damFire].Bolt		= new Array(0,27,81,82,83,84,85,2128);
    this.Info[damFire].Streak		= new Array(0,1796,1797,1798,1799,1800,1801,2129);
    this.Info[damFire].Volley		= new Array(0,0,0,143,144,145,146,0);
    this.Info[damFire].Blast		= new Array(0,0,0,115,116,117,118,0);
    this.Info[damFire].Arc		= new Array(0,2739,2740,2741,2742,2743,2744,2745);     
    this.Info[damFire].Wall		= 1841;
    this.Info[damFire].Ring		= 1785;
    this.Info[damFire].Rain		= 1834;
    this.Info[damFire].ProtSelf 	= new Array(0,20,1090,1091,1092,1093,1094,2157);
    this.Info[damFire].ProtOther	= new Array(0,19,810,836,849,1095,1096,2156);
    this.Info[damFire].Vuln		= new Array(0,21,1104,1105,1106,1107,1108,2170);
    this.Info[damFire].Bane		= new Array(0,1547,1548,1549,1550,1551,1552,2102);
    this.Info[damFire].Lure		= new Array(0,1541,1542,1543,1544,1545,1546,2103);

    this.Info[damAcid].Bolt		= new Array(0,58,59,60,61,62,63,2122);
    this.Info[damAcid].Streak		= new Array(0,1790,1791,1792,1793,1794,1795,2121);
    this.Info[damAcid].Volley		= new Array(0,0,0,127,128,129,130);
    this.Info[damAcid].Blast		= new Array(0,0,0,99,100,101,102,0);
    this.Info[damAcid].Arc		= new Array(0,2711,2712,2713,2714,2715,2716,2717);    
    this.Info[damAcid].Wall		= 1839;
    this.Info[damAcid].Ring		= 1783;
    this.Info[damAcid].Rain		= 1832;
    this.Info[damAcid].ProtSelf 	= new Array(0,515,516,517,518,519,520,2149);
    this.Info[damAcid].ProtOther	= new Array(0,509,510,511,512,513,514,2148);
    this.Info[damAcid].Vuln		= new Array(0,521,522,523,524,525,526,2162);
    this.Info[damAcid].Bane		= new Array(0,1493,1494,1495,1496,1497,1498,2092);
    this.Info[damAcid].Lure		= new Array(0,1499,1500,1501,1502,1503,1504,2093);  
    
    this.Info[damElectric].Bolt		= new Array(0,75,76,77,78,79,80,2140);
    this.Info[damElectric].Streak	= new Array(0,1814,1815,1816,1817,1818,1819,2141);
    this.Info[damElectric].Volley	= new Array(0,0,0,139,140,141,142,0);
    this.Info[damElectric].Blast	= new Array(0,0,0,111,112,113,114,0);
    this.Info[damElectric].Arc		= new Array(0,2732,2733,2734,2735,2736,2737,2738);     
    this.Info[damElectric].Wall		= 1844;
    this.Info[damElectric].Ring		= 1788;
    this.Info[damElectric].Rain		= 1837;
    this.Info[damElectric].ProtSelf 	= new Array(0,1066,1067,1068,1069,1070,1071,2159);
    this.Info[damElectric].ProtOther	= new Array(0,1072,1073,1074,1075,1076,1077,2158);
    this.Info[damElectric].Vuln		= new Array(0,1084,1085,1086,1087,1088,1089,2172);
    this.Info[damElectric].Bane		= new Array(0,1535,1536,1537,1538,1539,1540,2110);
    this.Info[damElectric].Lure		= new Array(0,1529,1530,1531,1532,1533,1534,2111);
    
    // There are no war spells that damage 'armor', so just list the life and items spells.
    this.Info[damArmor].ProtSelf 	= new Array(0,24,1308,1309,1310,1311,1312,2053);	// Armor Self
    this.Info[damArmor].ProtOther	= new Array(0,23,1313,1314,1315,1316,1317,2052);	// Armor Other
    this.Info[damArmor].Vuln		= new Array(0,25,1323,1324,1325,1326,1327,2074);	// Imperl
    this.Info[damArmor].Bane		= new Array(0,51,1482,1483,1484,1485,1486,2108);	// Impen
    this.Info[damArmor].Lure		= new Array(0,1487,1488,1489,1490,1491,1492,2100);	// Lure
    
    // Spell IDs for magic yield 1-7
    this.Yield				= new Array(0,280,281,282,283,284,285,2282);
    
    // Spell IDs for imperl 1-7
    this.Imperl				= new Array(0,25,1323,1324,1325,1326,1327,2074);
 
    // Percent increase in damage from vuln spells indexed by spell level
    this.VulnInc			= new Array(0,10,25,50,75,100,150,185);
    
    // War spells (bolt and arcs) minimum/maximum damages
    this.WarBoltMin			= new Array(0,8,13,18,31,46,61,110);
    this.WarBoltMax			= new Array(0,15,25,35,60,90,120,180);
    this.WarBoltRange			= new Array();
    for (var x=0; x<8; x++)
	this.WarBoltRange[x] = this.WarBoltMax[x] - this.WarBoltMin[x];
    
    // Calculate mininum damage levels for each vuln level by war spell level.
    // Note level 0 indicates damage values with no vuln and is just a copy of WarBoltMin
    this.MinDam	= new Array();
    for (var x=0; x<8; x++) {
        this.MinDam[x] = new Array();
        // Now for each war spell level, calculate the minimum damage
        for (var y=1; y<8; y++)
	    this.MinDam[x][y] = this.WarBoltMin[y] * (1+(this.VulnInc[y]/100));	
    } // endif
    
    // Methods
    this.Dump		= Damage_Dump;
    this.SetDebug	= Damage_SetDebug;
    
    return true;
}

function Damage_SetDebug(debugflag) {
    DamDebug = debugflag;
    return true;
}

// Used for debugging
function Damage_Dump() {

    for (x=1; x<9; x++) {
    
    	DamLog.OutD(" ");
	DamLog.OutD("Damage type [" + damName[x] + "]" );

	if (x<8) {
	    DamLog.OutD("Bolt Spells");
	    for (y=1; y<8; y++) {
	        var id = this.Info[x].Bolt[y];
	        if (id)
	            DamLog.OutD("Level: [" + y + "] Name   [" + Spell.Info[id].Name + "]");
	        else
	            DamLog.OutD("Level: [" + y + "] Undef");
	    } // endfor

	    DamLog.OutD("Streak Spells");
	    for (y=1; y<8; y++) {
	        var id = this.Info[x].Streak[y];
	        if (id)
	            DamLog.OutD("Level: [" + y + "] Name   [" + Spell.Info[id].Name + "]" );
	        else
	            DamLog.OutD("Level: [" + y + "] Undef" );
	    } // endfor
	
	    DamLog.OutD("Volley");	
	    for (y=1; y<8; y++) {
	        var id = this.Info[x].Volley[y];
	        if (id)
	            DamLog.OutD("Level: [" + y + "] Name   [" + Spell.Info[id].Name + "]");
	        else
	            DamLog.OutD("Level: [" + y + "] Undef");
	    } // endfor	
	
	    DamLog.OutD("Blast");	
	    for (y=1; y<8; y++) {
	        var id = this.Info[x].Blast[y];
	        if (id)
	            DamLog.OutD("Level: [" + y + "] Name   [" + Spell.Info[id].Name + "]");
	        else
	            DamLog.OutD("Level: [" + y + "] Undef");
	    } // endfor	
	
	    DamLog.OutD("Wall");
	    var id = this.Info[x].Wall;
	    if (id)
	        DamLog.OutD("Level: [" + y + "] Name   [" + Spell.Info[id].Name + "]");
	    else
	        DamLog.OutD("Level: [" + y + "] Undef");
		
	
	    DamLog.OutD ("Ring");
	    var id = this.Info[x].Ring;
	    if (id)
	        DamLog.OutD("Level: [" + y + "] Name   [" + Spell.Info[id].Name + "]");
	    else
	        DamLog.OutD("Level: [" + y + "] Undef");
		
	
	    DamLog.OutD ("Rain");
	    var id = this.Info[x].Rain;
	    if (id)
	        DamLog.OutD("Level: [" + y + "] Name   [" + Spell.Info[id].Name + "]");
	    else
	        DamLog.OutD("Level: [" + y + "] Undef");
	
	} // endif
		
	
	DamLog.OutD("SelfProts");
	for (y=1; y<8; y++) {
	    var id = this.Info[x].ProtSelf[y];
	    if (id)
	        DamLog.OutD("Level: [" + y + "] Name   [" + Spell.Info[id].Name + "]");
	    else
	        DamLog.OutD("Level: [" + y + "] Undef");
	} // endfor
	
	DamLog.OutD("OtherProts");
	for (y=1; y<8; y++) {
	    var id = this.Info[x].ProtOther[y];
	    if (id)
	        DamLog.OutD("Level: [" + y + "] Name   [" + Spell.Info[id].Name + "]");
	    else
	        DamLog.OutD("Level: [" + y + "] Undef");
	} // endfor
	
	DamLog.OutD("Vulns");	
	for (y=1; y<8; y++) {
	    var id = this.Info[x].Vuln[y];
	    if (id)
	        DamLog.OutD("Level: [" + y + "] Name   [" + Spell.Info[id].Name + "]");
	    else
	        DamLog.OutD("Level: [" + y + "] Undef");
	} // endfor
	
	DamLog.OutD("Banes");	
	for (y=1; y<8; y++) {
	    var id = this.Info[x].Bane[y];
	    if (id)
	        DamLog.OutD("Level: [" + y + "] Name   [" + Spell.Info[id].Name + "]");
	    else
	        DamLog.OutD("Level: [" + y + "] Undef");
	} // endfor
	
	DamLog.OutD("Lures");	
	for (y=1; y<8; y++) {
	    var id = this.Info[x].Lure[y];
	    if (id)
	        DamLog.OutD("Level: [" + y + "] Name   [" + Spell.Info[id].Name + "]");
	    else
	        DamLog.OutD("Level: [" + y + "] Undef");
	} // endfor
	
    } // endfor

} // end func

Damage = new DamageObj(false);