// MD_LOGOBJ.JS
//
// Author:	Poundboy
// Email:	poundboy@megapathdsl.net
// Version:	1.01
//
// Description:
//
// Collection of useful log/file functions/objects for Skunkworks.
//
//
// Modification History
//
// Who	Date	  Version	Description
//
// PB	07/07/02  1.01		Initial version
//
// Description:
//
// The following routines implement basic logging objects.
//
// LogObj will write output to the console and optionally to the chat area and/or a log file.
//
// LogFileObj will write lines of text to a disk file.
//
// The object contructors are as follows:
//
// myLogObj = new LogObj()
//
// myLogFileObj = new LogFileObj (filename, iomode, createflag)
//
//	filename	Filename to open
//	iomode		Access mode.  Valid modes: ForReading, ForWriting, ForAppending
//	createflag	Should we create file if it doesn't exist?
//
//
// The following methods are defined for LogObj:
//
// Out(text,chatflag)		Output a line of text.  If chatflag is true then also write line to chat area
// OutD(text,chatflag)		Output a line of text prefixed with the current date/time to current value of opm
// OutM(module,text,chatflag)	Output a line consisting of module+text fields.  If chatflag is true then also
//				write text field (not module name) to chat area
// OutDM(module,text,chatflag)	Output a line prefixed with the current date/time consisting of module+text
//				fields.  If chatflag is true just write text field (no module or date info) to chat area
// EnableLogFile(file)		Enable logging to file.  File opened in append mode.
// DisableLogFile()		Disables logging to file
//
// The following methods are defined for LogFileObj:
//
// Close()		Close log file
// Write(text)		Write text with no line terminator
// WriteD(text)		Write timestamp + text with no line terminator
// WriteLine(text)	Write text with line terminator
// WriteLineD(text)	Write timestamp + text with line terminator
//
// Requirements:
//
//    Requires the following libraries: MD_DEFS.JS, MD_UTIL.JS
//
// ****************************************************************************************
//
// Define a simple object for outputing text to the chat area, the skunkworks console, or
// to a log file
//
// Note since I define different log objects for each object class I define, I decided
// to have the log object check a global variable and if set output the log text
// to a file.  Note that any log object created will output text to a file if any one
// log object enables file logging.

// Global variable
LogFileEnabled	= false;

//
// Parameters:
//
// opm flag value for skapi.output call

function LogObj() {
    // Variables
    this.defColor	= cmcPaleBlue;
    // Methods
    this.Out		= Log_Output;
    this.OutD		= Log_OutputDate;
    this.OutM		= Log_OutputModule;
    this.OutDM		= Log_OutputModuleDate;
    this.EnableFile	= Log_EnableFile;
    this.DisableFile	= Log_DisableFile;
} // end LogObj

// Function for outputting log info to console and optionally file and/or chat area
function Log_Output(text, chatflag, chatcolor) {
    if ((chatflag == undefined) || (chatflag == null)) chatflag = false;
    if ((chatcolor == undefined) || (chatcolor == null)) chatcolor = this.defColor;

    skapi.OutputLine(text, opmConsole);
    if (chatflag) skapi.OutputLine(text, opmChatWnd, chatcolor);
    if (LogFileEnabled) LogFileX.WriteLine(text);
} // end function

// Function for outputting log info to console and optionally file and/or chat area with timestamp prefix
function Log_OutputDate(text, chatflag, chatcolor) {
    if ((chatflag == undefined) || (chatflag == null)) chatflag = false;
    if ((chatcolor == undefined) || (chatcolor == null)) chatcolor = this.defColor;

    skapi.OutputLine(Util.FormatDate() + " " + text, opmConsole);
    if (chatflag) skapi.OutputLine(text, opmChatWnd, chatcolor);
    if (LogFileEnabled) LogFileX.WriteLineD(text);
} // end function

// Function for outputting log info to console and optionally file and/or chat area
function  Log_OutputModule(module, text, chatflag, chatcolor) {
    if ((chatflag == undefined) || (chatflag == null)) chatflag = false;
    if ((chatcolor == undefined) || (chatcolor == null)) chatcolor = this.defColor;

    skapi.OutputLine(module + ": " + text, opmConsole);
    if (chatflag) skapi.OutputLine(text, opmChatWnd, chatcolor);
    if (LogFileEnabled) LogFileX.WriteLine(module + ": " + text);
} // end function

// Function for outputting log info to console and optionally file and/or chat area with timestamp prefix
function Log_OutputModuleDate(module, text, chatflag, chatcolor) {
    if ((chatflag == undefined) || (chatflag == null)) chatflag = false;
    if ((chatcolor == undefined) || (chatcolor == null)) chatcolor = this.defColor;

    skapi.OutputLine(Util.FormatDate() + " " + module + ": " + text, opmConsole);
    if (chatflag) skapi.OutputLine(text, opmChatWnd, chatcolor);
    if (LogFileEnabled) LogFileX.WriteLineD(module + ": " + text);
} // end function

function Log_EnableFile(FileName) {
    LogFileEnabled = true;
    LogFileX 	   = new LogFileObj(FileName, ForAppending, true);
} // end function

function Log_DisableFile() {
    LogFileEnabled = false;
} // end function

// *******************************************************************************************
// Implement a simple log file object for writing text to a file.  No read functions supported
//
// Parameters:
//
// filename		Filename to open
// iomode		Access mode.  Valid modes: ForReading, ForWriting, ForAppending
// create_flag		Should we create file if it doesn't exist?
//
// Returns:
//
// TextStream object

function LogFileObj(filename, iomode, create_flag) {
    this.myFileSysObj	= new ActiveXObject("Scripting.FileSystemObject");
    this.TextFileObj	= this.myFileSysObj.OpenTextFile(filename, iomode, create_flag);

    // Implement a subset of the TextStream object methods
    this.Close		= LogFile_Close;	// Close log file
    this.Write		= LogFile_Write;	// Write text with no line terminator
    this.WriteD		= LogFile_WriteD;	// Write timestamp + text with no line terminator
    this.WriteLine	= LogFile_WriteLine;	// Write text with line terminator
    this.WriteLineD	= LogFile_WriteLineD;	// Write timestamp + text with line terminator
}

function LogFile_Close() {
    return this.TextFileObj.Close();
}

function LogFile_Write(text) {
    return this.TextFileObj.Write(text);
}

function LogFile_WriteD(text) {
    return this.TextFileObj.Write(Util_FormatDate() + " " + text);
}

function LogFile_WriteLine(text) {
    return this.TextFileObj.WriteLine(text);
}

function LogFile_WriteLineD(text) {
    return this.TextFileObj.WriteLine(Util_FormatDate() + " " + text);
}