// MD_PLAYEROBJ.JS
//
// Author:	Poundboy
// Email:	poundboy@megapathdsl.net
// Version:	1.01
//
// Description:
//
// Collection of player based functions/objects for Skunkworks. 
// Note!!!  All pixel offsets based on 1024x768 screen,
//
// Requirements:
//
//    Requires the following libraries:	MD_UTIL.JS, MD_LOGOBJ.JS
//
// Methods:
//
//  EnsureCombatMode(boolean)	If boolean!=skapi.fCombatMode, then toggle mode key.
//
//  CreateFellow(name)		Create fellowship with name 'name'.
//
//  AddFellow(name)		Add player 'name' to our fellowship
//
//  DisbandFellow()		If we are the leader, then disband the fellowship
//
//  DropFellow()		If we are not the leader, then drop out of fellow
//
//  MemberFellow()		Returns true if part of fellowship, false otherwise
//
//  VerifyFellowMbr(name)	Returns true if 'name' is a member of our fellow, false otherwise					
//	
//  LogOff()			Logoff character.
//	
//  LogOn(char#)		Logon character.  Char# of numbered 0-4.
//
//  ToSlotNum(name)		Return character 'name' login slot #.  Slots numbered from 0-4.
//
//  CreatureCount(radius)	Returns # of creatures within 'radius' map units of us
//
//  DangerCheck(radius,#cre)	Returns TRUE if '#cre' creatures within 'radius' map units
//
//  SetDebug(boolean,opm)	If set to true, will log debug information to opm destination.
//
//  SetSwitchWait(value)	Time to wait before assuming combat mode toggle failed. Default 4000ms
//
//  SetModeWait(value)		Time to wait after successful combat mode switch.  Ensures
//				animation has time to complete.
//
//  SetFellowInt(value)		Delay time between fellowship creation actions. Default is 100ms.
//				Increase if fellowship creation fails because actions are happening
//				too fast.
//
//  SetLogoutTimeout(value)	Sets time (ms) to wait for logout to complete before giving up.
//				Default is 40 seconds.	
//
//  SetLogonTimeout(value)	Sets time (ms) to wait for logon to complete before giving up.
//				Default is 40 seconds.
//
//
// Modification History
//
// Who	Date	  Version	Description
//
// PBoy	07/07/02  1.01		Initial version
//


// Global variables/constants
PlayerLog	= new LogObj();		// Log object used to output to console/chat window
PlayerDebug	= false;		// If set to true, will output debugging information

// Set up skunkworks event handler structures
function PlayerHandlers_OnCombatMode(fOn) {
    if (PlayerDebug) Log.OutD("Player_OnCombatMode ["+ fOn + "]");
    OnCombatFired = true;
    OnCombatValue = fOn;
} // end function

function PlayerHandlers() {
    this.OnCombatMode = PlayerHandlers_OnCombatMode;
} // end func

var PlayerH	  = new PlayerHandlers();	// Set up handlers variable as a global variable
var PlayerHandReg = false;			// Set to true when handlers registered


// **************************************************************************************************
// Player Object

function PlayerObj(debugflg) {
    if ((debugflg==undefined) || (debugflg==null)) debugflg = false;
    PlayerDebug = debugflg;	// Outputs debugging info to debug destination if set

    // Properties	
    this.LogoutTimeout	= 40000;	// Default time (ms) to wait for player logoff.
    this.LogonTimeout	= 40000;	// Default time (ms) to wait for player logon.
    this.FellowInt	= 300;		// Interval to wait between fellow actions
    this.SwitchWait	= 4000;		// Time (ms) to wait for mode switch before aborting
    this.ModeWait	= 250;		// Time (ms) to wait after mode switch for animation to
    					// complete
    					
    this.CreOid		= 0;		// Oid of creature found with CreatureCount method if only
    					// one creature found.  
    					  
    // Methods
    this.SetDebug		= Player_SetDebug;
    this.SetSwitchWait		= Player_SetSwitchWait;
    this.SetModeWait		= Player_SetModeWait;
    this.SetFellowInt		= Player_SetFellowInt;
    this.SetLogoutTimeout	= Player_SetLogoutTimeout;
    this.SetLogonTimeout	= Player_SetLogonTimeout;  
    this.EnsureCombatMode	= Player_EnsureCombatMode;
    this.CreateFellow		= Player_CreateFellow;
    this.AddFellow		= Player_AddFellow;
    this.DisbandFellow		= Player_DisbandFellow;
    this.DropFellow		= Player_DropFellow;
    this.MemberFellow		= Player_MemberFellow;
    this.ToggleFellow		= Player_ToggleFellow;
    this.VerifyFellowMbr	= Player_VerifyFellowMbr;
    this.LogOff			= Player_LogOff;
    this.LogOn			= Player_LogOn;
    this.ToSlotNum		= Player_ToSlotNum;
    this.CreatureCount		= Player_CreatureCount;
    this.DangerCheck		= Player_DangerCheck;
    
    if (!PlayerHandReg)	{
	skapi.AddHandler( evidOnCombatMode, PlayerH );
	PlayerHandReg = true;
    } // endif
} // end func

function Player_SetDebug(debugflag) {
    PlayerDebug = debugflag;
    return true;
}

function Player_SetSwitchWait(value) {
    this.SwitchWait = value;
}

function Player_SetModeWait(value) {
    this.ModeWait = value;
}

function Player_SetFellowInt(value) {
    this.FellowInt = value;
}

function Player_SetLogoutTimeout(value) {
    this.LogoutTimeout = value;
}

function Player_SetLogonTimeout(value) {
    this.LogonTimeout = value;
}

function Player_EnsureCombatMode(combatmode) {
    if (PlayerDebug)
	PlayerLog.OutD ("Player_EnsureCombatMode: Begin, combatmode [" + combatmode + "]");
    
    // If already in correct mode, then exit
    if (skapi.fCombatMode==combatmode) {
    	if (PlayerDebug) PlayerLog.OutD("Player_EnsureCombatMode: End, No mode change necessary");
    	return;
    } // endif
    
    OnCombatFired = false;
     
    var SwitchStart = new Date();
    if (PlayerDebug) PlayerLog.OutD("Player_EnsureCombatMode: Setting mode...");
    skapi.SetCombatMode(combatmode);
          
    // Now wait for mode to change
    if (PlayerDebug) PlayerLog.OutD("Player_EnsureCombatMode: Waiting for event or timeout...");
    while (((new Date()-SwitchStart) < this.SwitchWait) && !OnCombatFired) {
	skapi.WaitEvent(20,wemFullTimeout);
    } // end while
  
    // If couldn't switch mode, report error
    if (PlayerDebug) PlayerLog.OutD("Player_EnsureCombatMode: Wait done.");
    if (combatmode != skapi.fCombatMode) {
	if (PlayerDebug) PlayerLog.OutD("Player_EnsureCombatMode: End, Failure to switch modes");
        return false;
    } // endif
     
    // Mode switched successfully, now wait a little bit more to make sure animation complete
    if (PlayerDebug)
	PlayerLog.OutD("Player_EnsureCombatMode: Waiting [" + this.ModeWait +
		       "] ms for animation to complete...");
    skapi.WaitEvent(this.ModeWait, wemFullTimeout);

    if (PlayerDebug) PlayerLog.OutD("Player_EnsureCombatMode: End");
    return true;
} // end func Player_EnsureCombatMode


function Player_CreateFellow(name) {
    // open map attribute panel to ensure fellowship panel not displayed
    skapi.Keys(cmidAttributesPanel, kmoClick);
    skapi.sleep(this.FellowInt);
	
    // open fellowship panel
    skapi.Keys(cmidFellowshipPanel, kmoClick);
    skapi.sleep(this.FellowInt);
	
    // position cursor in fellowship name area
    skapi.MouseMove(skapi.dxpRes-20, 262);
    skapi.sleep(this.FellowInt);
	
    // now click to put focus on fellowship name box
    skapi.MouseButton(ibuttonLeft, kmoClick);
    skapi.sleep(this.FellowInt);

    // Now it backspace several times to delete any current name in box
    for (var x=0; x<30; x++) {
	skapi.Keys("{back}", kmoClick);
	skapi.sleep(20);
    } //endif

    // Enter fellowship name
    skapi.PasteSz(name);
    skapi.sleep(this.FellowInt);
	
    // Position cursor to create fellowship button and click
    skapi.MouseMove(skapi.dxpRes-150, 367);
    skapi.sleep(this.FellowInt);
    skapi.MouseButton(ibuttonLeft, kmoClick);
    skapi.sleep(this.FellowInt);	
} // end function


function Player_AddFellow(name) {
    PlayerLog.OutD("Player_AddFellow: Begin, Name [" + name + "]");
    var FellowAco = skapi.AcoFromSz(name);
    if (FellowAco == null) {
    	PlayerLog.OutD("Player_AddFellow: Unable to get ACO for player");
    	return false;
    } // endif
  
    // Now make sure player near us so we don't wander off trying to fellow him/her 
    var OurLoc = skapi.maplocCur;
    var FellowLoc = FellowAco.maploc;
    
    var Dist  = Util.Distance(FellowLoc,OurLoc);
    if (Dist>0.10) {
    	PlayerLog.OutD("Player_AddFellow: Player too far away");
    	Chat.Tell(name, "Failed: You are " + Util.Round(Dist,3) +
    		        " map units away. Come within 0.10 units");
    	return false;
    } // endif
    
    // Open attribute panel to ensure fellowship panel not displayed
    skapi.Keys(cmidAttributesPanel, kmoClick);
    skapi.sleep(this.FellowInt);
	
    // Now open fellowship panel
    skapi.Keys(cmidFellowshipPanel, kmoClick);
    skapi.sleep(this.FellowInt);
    
    // We have the ACO of the player we want to fellow and he/she is close enough to us.
    // Now try to select the player up to 3 times.
    PlayerLog.OutD("Player_AddFellow: Attempting to select fellow"); 
    for (var x=0; x<3; x++) {
    	// Select player
    	skapi.SelectAco(FellowAco);
    	skapi.WaitEvent(200, wemFullTimeout); // Pause a little bit
    	if (skapi.acoSelected == FellowAco) break;
    } // endfor
    
    if (skapi.acoSelected != FellowAco) {
    	PlayerLog.OutD("Player_AddFellow: Could not select fellow char"); 
    	return false;
    } // endif

    // Now position cursor to recruit button and click
    PlayerLog.OutD ("Fellowing player...");
    skapi.MouseMove(skapi.dxpRes-243, skapi.dypRes-108);	// 781, 660 for 1024x768
    skapi.sleep(this.FellowInt);
    skapi.MouseButton(ibuttonLeft, kmoClick);
    skapi.sleep(this.FellowInt);	
    return true;	// Return success.  In future check to see if player was really added
    
} // end function

function Player_DisbandFellow() {
    // Open attributes panel to ensure fellowship panel not displayed
    skapi.Keys(cmidAttributesPanel, kmoClick);
    skapi.sleep(this.FellowInt);
	
    // Now open fellowship panel
    skapi.Keys(cmidFellowshipPanel, kmoClick);
    skapi.sleep(this.FellowInt);
    
    // Now position cursor to disband button and click
    PlayerLog.OutD ("Disbanding fellowship...");
    skapi.MouseMove(skapi.dxpRes-63, skapi.dypRes-108);		// 961, 660
    skapi.sleep(this.FellowInt);
    skapi.MouseButton(ibuttonLeft, kmoClick);
    skapi.sleep(this.FellowInt);	
} // end function Player_DisbandFellow


function Player_DropFellow() {
    // Open attributes panel to ensure fellowship panel not displayed
    skapi.Keys(cmidAttributesPanel, kmoClick);
    skapi.sleep(this.FellowInt);
	
    // Now open fellowship panel
    skapi.Keys(cmidFellowshipPanel, kmoClick);
    skapi.sleep(this.FellowInt);
    
    // Now position cursor to Quit Fellowship button and click
    PlayerLog.OutD ("Dropping out of fellowship...");
    skapi.MouseMove(skapi.dxpRes-153, skapi.dypRes-141);	// 871,627
    skapi.sleep(this.FellowInt);
    skapi.MouseButton(ibuttonLeft, kmoClick);
    skapi.sleep(this.FellowInt);	
} // end function Player_DropFellow

function Player_MemberFellow() {
    if ((skapi.szFellowship == undefined) || (skapi.szFellowship == null) || (skapi.szFellowship == "")) return false;
    var cofellow = skapi.cofellow;
    if (cofellow.Count == 0) return false;
    return true;
} // endif

// Toggle OPEN/CLOSE status of fellow
function Player_ToggleFellow() {
    // Open attributes panel to ensure fellowship panel not displayed
    skapi.Keys(cmidAttributesPanel, kmoClick);
    skapi.sleep(this.FellowInt);
	
    // Now open fellowship panel
    skapi.Keys(cmidFellowshipPanel, kmoClick);
    skapi.sleep(this.FellowInt);
    
    // Now position cursor to OPEN/CLOSE button and click
    PlayerLog.OutD("Toggling fellowship OPEN/CLOSE status...");
    skapi.MouseMove(skapi.dxpRes-63, skapi.dypRes-141);		// 961, 627
    skapi.sleep(this.FellowInt);
    skapi.MouseButton(ibuttonLeft, kmoClick);
    skapi.sleep(this.FellowInt);	  
} // endif

function Player_VerifyFellowMbr(name) {
    // Gather fellow information
    var cofellow = skapi.cofellow;
    
    // If we are not past of a fellow, then return FALSE
    if (cofellow.Count == 0) return false;
    
    // See if player 'name' is a member of our fellow. Return TRUE if so.
    for (var x=0; x<cofellow.Count; x++) {
    	var FellowAco = cofellow.Item(x);
    	if (FellowAco.szName == name) return true;
    } // endfor
    
    // Player not found in fellow list, not a member of our fellow
    return false;
} // endif
	   

function Player_LogOff() {
    if (PlayerDebug) PlayerLog.OutD("Player_LogOff: Begin");

    if (PlayerDebug) PlayerLog.OutD("Player_LogOff: Sending logout shortcut...");
    var LogoffStart = new Date();	// Init timer for logoff attempt
    
    // Old method
    // skapi.Keys("~{ESC}", kmoClick);	// Send logout shortcut
    // skapi.Sleep(2000);
    // skapi.Keys("~{ESC}", kmoClick);	// Send logout shortcut 2nd time in case 1st one failed
    
    // Improved method
    skapi.Logout();

    // Now make sure we are completely logged out
    if (PlayerDebug) PlayerLog.OutD ("Player_LogOff: Waiting for char to logout...");
    while ((skapi.plig != pligAtLogin) && ((new Date()-LogoffStart) < this.LogoutTimeout))
   	skapi.WaitEvent(50);
   	
    if (PlayerDebug) PlayerLog.OutD("Player_LogOff: AtLogin wait done");
	
    if (skapi.plig == pligAtLogin) {
	if (PlayerDebug) PlayerLog.OutD("Player_Logoff: End, Toon successfully logged out");
	return true;
    } // endif
	
    if (PlayerDebug) PlayerLog.OutD("Player_LogOff: End, Timeout" );
    return false;
	
} // end func

function Player_LogOn(char_slot) {
    if (PlayerDebug) PlayerLog.OutD ("Player_LogOn: Begin");
	
    if ((char_slot<0) || (char_slot>6)) {
	PlayerLog.OutD( "Player_LogOn: Character slot must be from 0 to 6");
	return false;
    } // endif
	
    var LogonStart = new Date();		// event when completely logged on.
	
    // Compute absolute x and y screen coords of slots name
    var xOffset = 100;
    var yOffset = 222;
    yOffset += (char_slot * 45);
   	
    // Now move mouse to coordinates of login chars name
    if (PlayerDebug)
        PlayerLog.OutD("Player_LogOn: Moving mouse to coords (" + xOffset + "," + yOffset + ")");
    skapi.MouseMove(xOffset, yOffset);
    skapi.Sleep(400);
    
    // and double click
    skapi.MouseButton(ibuttonLeft, kmoClick);
    skapi.Sleep(200);
    skapi.MouseButton(ibuttonLeft, kmoClick);
    skapi.Sleep(200);
    
    // Move over to Enter button and click it. 
    skapi.MouseMove(350,400);
    skapi.Sleep(200);
    skapi.MouseButton(ibuttonLeft, kmoClick);
    
    // One more time just in case
    skapi.MouseMove(350,400);
    skapi.Sleep(200);
    skapi.MouseButton(ibuttonLeft, kmoClick);
	    
    // Now make sure we are completely logged in
    while ((skapi.plig != pligInWorld) && ((new Date()-LogonStart) < this.LogonTimeout))
   	skapi.WaitEvent(50);
    if (PlayerDebug) PlayerLog.OutD("Player_LogOn: InWorld wait done");

    if (skapi.plig == pligInWorld) {
	if (PlayerDebug) PlayerLog.OutD("Player_LogOn: End, Toon successfully logged on");
	return true;
    }
		
    if (PlayerDebug)
	PlayerLog.OutD("Player_LogOn: End, Error logging on toon, state [" + skapi.plig + "]");
    return false;
	
    if (PlayerDebug) PlayerLog.OutD("Player_LogOn: End");
} // end func

function Player_ToSlotNum(name) {
    if (PlayerDebug) PlayerLog.OutD("Player_ToSlotNum: Begin, Name [" + name + "]");
    
    for (var x=0; x<skapi.cchar; x++)
        if (name.toUpperCase() == skapi.rgszChar(x).toUpperCase()) {
            if (PlayerDebug) PlayerLog.OutD("Player_ToSlotNum: End, Match found [" + x + "]");
	    return x;
	} // endif
    
    if (PlayerDebug) PlayerLog.OutD("Player_ToSlotNum: End, Name unknown");
    return -1; // No match found
    
} // end func

 
function Player_CreatureCount(Radius) {
    if (PlayerDebug)
	PlayerLog.OutD("Player_CreatureCount: Begin, Radius [" + Radius + "]");
    
      // Create aco collection filter
    var acf = skapi.AcfNew();
    acf.Reset();
    acf.ocm = ocmMonster;
    acf.distMax = Radius;
    var MonList = acf.CoacoGet();
    
    // If only one creature found, then set CreOid with it's oid
    if (MonList.Count == 1) {
    	MonAco = MonList.Item(0);
    	this.CreOid = MonAco.oid;
    } else {
    	this.CreOid = 0;
    } // endif
    
    if (PlayerDebug)
	PlayerLog.OutD("Player_CreatureCount: End, # of creatures found [" + MonList.Count + "]");
	
    return MonList.Count;
} // end function
 

function Player_DangerCheck(Radius,Thresh) {
    if (PlayerDebug)
	PlayerLog.OutD("Player_DangerCheck: Begin, Radius [" + Radius + "] Thresh [" + Thresh +"]");
    if ((Thresh==undefined) || (Thresh==null)) Thresh = 1;
    
      // Create aco collection filter
    var acf = skapi.AcfNew();
    acf.Reset();
    acf.ocm = ocmMonster;
    acf.distMax = Radius;
    var MonList = acf.CoacoGet();
    
    if (PlayerDebug)
	PlayerLog.OutD("Player_DangerCheck: # of creatures found [" + MonList.Count + "]");
	
    if (MonList.Count < Thresh) {
	if (PlayerDebug) PlayerLog.OutD("Player_DangerCheck: End, Return value: FALSE");
	return false;
    } // endif
    
    if (PlayerDebug) PlayerLog.OutD("Player_DangerCheck: End, Return value: TRUE");
    return true; 
} // end function

// Create player object 
Player = new PlayerObj(false);